from .base import (SupportsRecordsDirectory, SupportsMoveToRecordsDirectory,  # noqa
                   SupportsToFileobjsSource, RecordsSource, SupportsToDataframesSource)
from .fileobjs import FileobjsSource  # noqa
from .factory import RecordsSources  # noqa
from .dataframes import DataframesRecordsSource  # noqa



from typing import Callable, Optional, Union
from sqlalchemy import MetaData
from sqlalchemy.engine import Engine, Connection
from .processing_instructions import ProcessingInstructions
from .existing_table_handling import ExistingTableHandling
import logging
from logging import Logger
from .unload_plan import RecordsUnloadPlan
from .records_format import RecordsFormat
from ..db import DBDriver, LoadError
from ..url.resolver import UrlResolver
from .mover import move
from .types import RecordsFormatType
from .sources import RecordsSources
from .targets import RecordsTargets

logger = logging.getLogger(__name__)


class Records:
    def __init__(self,
                 db_driver: Callable[[Union[Engine, Connection]], DBDriver],
                 url_resolver: UrlResolver,
                 logger: Logger=logger) -> None:
        self.meta = MetaData()
        self.db_driver = db_driver
        self.url_resolver = url_resolver
        self.move = move
        self.RecordsFormat = RecordsFormat
        self.RecordsUnloadPlan = RecordsUnloadPlan
        self.ProcessingInstructions = ProcessingInstructions
        self.ExistingTableHandling = ExistingTableHandling
        self.LoadError = LoadError
        self.logger = logger
        self.sources = RecordsSources(db_driver=db_driver,
                                      url_resolver=url_resolver)
        self.targets = RecordsTargets(url_resolver=url_resolver,
                                      db_driver=db_driver)
        """
        To move records from one place to another, you can use the methods on this object.

        Example:

        .. code-block:: python

           records = job_context.records
           db_engine = job_context.get_default_db_engine()
           url = 's3://some-bucket/some-directory/'
           source = records.sources.directory_from_url(url=url)
           target = records.targets.table(schema_name='myschema',
                                          table_name='mytable',
                                          db_engine=db_engine)
           results = records.move(source, target)
        """

    def best_records_format_variant(self,
                                    records_format_type: RecordsFormatType,
                                    db_engine: Engine) -> Optional[str]:
        driver = self.db_driver(db_engine)
        return driver.best_records_format_variant(records_format_type)



from contextlib import contextmanager

from .cli import cli_logging
from . import log_levels
from .cli.cli_job_context import CLIJobContext
from .itest_job_context import IntegrationTestJobContext
from .env_job_context import EnvJobContext
from .airflow.airflow_job_context import AirflowJobContext
from .temp_dir import set_temp_dir
from .base_job_context import BaseJobContext, JsonSchema
import atexit
from typing import Iterator, Optional, Type
import os


@contextmanager
def create_job_context(name: str, **kwargs) -> Iterator[BaseJobContext]:
    """Use this to run a job.  It will yield a job context, which you can use
    to get information about your job's place in the world!
    """
    jc = __get_job_context(name, **kwargs)
    try:
        with set_temp_dir(jc.temp_dir):
            yield jc
    finally:
        jc.cleanup()


def pull_job_context(name: str, **kwargs) -> BaseJobContext:
    """Returns a job context.  Cleans up temporary resources at process
    exit time, rather than as a context manager - see
    create_job_context for that.
    """
    job_context_contextmanager = create_job_context(name, **kwargs)
    atexit.register(lambda: job_context_contextmanager.__exit__(None, None, None))
    return job_context_contextmanager.__enter__()


def run_as_job(fn, name: str, **kwargs) -> None:
    with create_job_context(name, **kwargs) as job_context:
        fn(job_context)


def __get_job_context(name: str,
                      use_default_logging: bool=True,
                      default_db_creds_name: Optional[str]=None,
                      default_aws_creds_name: Optional[str]=None,
                      scratch_s3_url: Optional[str]=None,
                      config_json_schema: JsonSchema=None,
                      job_context_type: Optional[str]=None) -> BaseJobContext:
    if job_context_type is None:
        job_context_type = os.environ.get('PY_JOB_CONTEXT')
    log_levels.set_levels(name)
    jc_class: Type[BaseJobContext]
    if job_context_type == 'airflow':
        jc_class = AirflowJobContext
    elif job_context_type == 'cli':
        jc_class = CLIJobContext
    elif job_context_type == 'itest':
        jc_class = IntegrationTestJobContext
    elif job_context_type == 'env':
        jc_class = EnvJobContext
    elif job_context_type is not None:
        raise ValueError("Valid job context types: cli, airflow, docker-itest, env - "
                         "consider upgrading records-mover if you're looking for "
                         f"{job_context_type}.")
    elif 'AIRFLOW__CORE__EXECUTOR' in os.environ:
        # Guess based on an env variable sometimes set by Airflow
        jc_class = AirflowJobContext
    else:
        jc_class = CLIJobContext

    if jc_class != AirflowJobContext:
        if use_default_logging:
            cli_logging.basic_config()

    return jc_class(name,
                    config_json_schema=config_json_schema,
                    default_db_creds_name=default_db_creds_name,
                    default_aws_creds_name=default_aws_creds_name,
                    scratch_s3_url=scratch_s3_url)



from .constraints import RecordsSchemaFieldConstraints
import sqlalchemy
from typing import Optional, cast, TYPE_CHECKING
if TYPE_CHECKING:
    from .constraints import FieldDecimalConstraintsDict  # noqa
    from records_mover.db import DBDriver  # noqa


class RecordsSchemaFieldDecimalConstraints(RecordsSchemaFieldConstraints):
    def __init__(self,
                 required: bool,
                 unique: Optional[bool],
                 fixed_precision: Optional[int]=None,
                 fixed_scale: Optional[int]=None,
                 fp_total_bits: Optional[int]=None,
                 fp_significand_bits: Optional[int]=None):
        super().__init__(required=required, unique=unique)
        self.fixed_precision = fixed_precision
        self.fixed_scale = fixed_scale
        self.fp_total_bits = fp_total_bits
        self.fp_significand_bits = fp_significand_bits

    @staticmethod
    def from_sqlalchemy_type(required: bool,
                             unique: Optional[bool],
                             type_: sqlalchemy.types.TypeEngine,
                             driver: 'DBDriver') -> 'RecordsSchemaFieldDecimalConstraints':
        if not isinstance(type_, sqlalchemy.types.Numeric):
            raise TypeError(f"Unexpected column type: {type_}")

        if isinstance(type_, sqlalchemy.types.Float):
            constraints = driver.fp_constraints(type_)
            fp_total_bits: Optional[int] = None
            fp_significand_bits: Optional[int] = None
            if constraints:
                fp_total_bits, fp_significand_bits = constraints

            return RecordsSchemaFieldDecimalConstraints(required=required,
                                                        unique=unique,
                                                        fp_total_bits=fp_total_bits,
                                                        fp_significand_bits=fp_significand_bits)
        else:
            constraints = driver.fixed_point_constraints(type_)
            fixed_precision: Optional[int] = None
            fixed_scale: Optional[int] = None
            if constraints:
                fixed_precision, fixed_scale = constraints

            return RecordsSchemaFieldDecimalConstraints(required=required,
                                                        unique=unique,
                                                        fixed_precision=fixed_precision,
                                                        fixed_scale=fixed_scale)

    def to_data(self) -> 'FieldDecimalConstraintsDict':
        raw_out = super().to_data()
        out = cast('FieldDecimalConstraintsDict', raw_out)
        if self.fixed_precision is not None:
            out['fixed_precision'] = self.fixed_precision
        if self.fixed_scale is not None:
            out['fixed_scale'] = self.fixed_scale
        if self.fp_total_bits is not None:
            out['fp_total_bits'] = self.fp_total_bits
        if self.fp_significand_bits is not None:
            out['fp_significand_bits'] = self.fp_significand_bits
        return out



import logging
from sqlalchemy import Column
import sqlalchemy
from typing import cast, Union, Optional, Type, TYPE_CHECKING
from .constraints import (RecordsSchemaFieldConstraints,
                          RecordsSchemaFieldIntegerConstraints,
                          RecordsSchemaFieldDecimalConstraints,
                          RecordsSchemaFieldStringConstraints)
from .representation import RecordsSchemaFieldRepresentation
from .statistics import RecordsSchemaFieldStringStatistics
from .string_length_generator import generate_string_length
if TYPE_CHECKING:
    from ....db import DBDriver  # noqa
    from ..field import RecordsSchemaField  # noqa
    from ..schema import RecordsSchema  # noqa
    from .types import FieldType  # noqa


logger = logging.getLogger(__name__)


def field_from_sqlalchemy_column(column: Column,
                                 driver: 'DBDriver',
                                 rep_type: str)\
            -> 'RecordsSchemaField':
    from ..field import RecordsSchemaField  # noqa
    name = column.name
    type_: Union[Type[sqlalchemy.types.TypeEngine],
                 sqlalchemy.types.TypeEngine] = column.type
    constraints: Optional[RecordsSchemaFieldConstraints] = None
    field_type = None  # type: Optional[FieldType]

    required = not column.nullable

    # We don't yet try to pull information about uniqueness
    # constraints out of database tables, so set this to null,
    # which means "we don't know":
    #
    # https://app.asana.com/0/1128138765527694/1131416227825120
    unique = None

    if isinstance(type_, sqlalchemy.sql.sqltypes.Integer):
        field_type = 'integer'
        constraints = RecordsSchemaFieldIntegerConstraints.\
            from_sqlalchemy_type(required=required,
                                 unique=unique,
                                 type_=type_,
                                 driver=driver)
    elif (isinstance(type_, (sqlalchemy.sql.sqltypes.Numeric,
                             sqlalchemy.sql.sqltypes.Float,
                             sqlalchemy.dialects.postgresql.base.DOUBLE_PRECISION))):
        field_type = 'decimal'
        constraints = RecordsSchemaFieldDecimalConstraints.\
            from_sqlalchemy_type(required=required,
                                 unique=unique,
                                 type_=type_,
                                 driver=driver)
    elif (isinstance(type_, sqlalchemy.sql.sqltypes.String)):
        field_type = 'string'
        constraints = RecordsSchemaFieldStringConstraints.\
            from_sqlalchemy_type(required=required,
                                 unique=unique,
                                 type_=type_,
                                 driver=driver)
    elif isinstance(type_, sqlalchemy.sql.sqltypes.Date):
        field_type = 'date'
    elif isinstance(type_, sqlalchemy.sql.sqltypes.DateTime):
        date_plus_time_with_timezone = driver.type_for_date_plus_time(has_tz=True)
        date_plus_time_no_timezone = driver.type_for_date_plus_time(has_tz=False)

        #
        # Databases are super annoying about these.
        #
        # BigQuery's SQLAlchemy driver uses TIMEZONE for the
        # former, DATETIME for the latter, without setting the
        # 'timezone' variable.
        #
        # Redshift and Vertica use DateTime for both with the
        # 'timezone' variable set appropriately.
        #
        if type(date_plus_time_with_timezone) != type(date_plus_time_no_timezone):
            if isinstance(type_, type(date_plus_time_with_timezone)):
                field_type = 'datetimetz'
            elif isinstance(type_, type(date_plus_time_no_timezone)):
                field_type = 'datetime'
            else:
                raise NotImplementedError(f"Teach me how to handle {type_}")
        else:
            if type_.timezone:
                field_type = 'datetimetz'
            else:
                field_type = 'datetime'
    elif isinstance(type_, sqlalchemy.sql.sqltypes.Time):
        if type_.timezone:
            field_type = 'timetz'
        else:
            field_type = 'time'
    elif isinstance(type_, sqlalchemy.sql.sqltypes.Boolean):
        field_type = 'boolean'
    else:
        raise NotImplementedError("Teach me how to handle this SQLAlchemy type: "
                                  f"{type(type_)}")
    if constraints is None:
        constraints = RecordsSchemaFieldConstraints.\
            from_sqlalchemy_type(required=required,
                                 unique=unique,
                                 type_=type_,
                                 driver=driver)
    representations = {
        'origin': RecordsSchemaFieldRepresentation.
        from_sqlalchemy_column(column, driver.db.dialect,
                               rep_type)
    }

    # We don't currently gather statistics from databases - which
    # can bite us when exporting from BigQuery, for instance:
    #
    # https://app.asana.com/0/53283930106309/1131698268455053
    statistics = None

    return RecordsSchemaField(name=name,
                              field_type=field_type,
                              constraints=constraints,
                              statistics=statistics,
                              representations=representations)


def field_to_sqlalchemy_type(field: 'RecordsSchemaField',
                             driver: 'DBDriver') -> sqlalchemy.types.TypeEngine:
    if field.field_type == 'integer':
        int_constraints =\
            cast(Optional[RecordsSchemaFieldIntegerConstraints], field.constraints)
        min_: Optional[int] = None
        max_: Optional[int] = None
        if int_constraints:
            min_ = int_constraints.min_
            max_ = int_constraints.max_
        return driver.type_for_integer(min_value=min_, max_value=max_)
    elif field.field_type == 'decimal':
        decimal_constraints =\
            cast(Optional[RecordsSchemaFieldDecimalConstraints], field.constraints)
        if decimal_constraints:
            if (decimal_constraints.fixed_precision is not None and
               decimal_constraints.fixed_scale is not None):
                return driver.type_for_fixed_point(precision=decimal_constraints.fixed_precision,
                                                   scale=decimal_constraints.fixed_scale)

            if (decimal_constraints.fp_total_bits is not None and
               decimal_constraints.fp_significand_bits is not None):
                return driver.\
                    type_for_floating_point(fp_total_bits=decimal_constraints.fp_total_bits,
                                            fp_significand_bits=decimal_constraints.
                                            fp_significand_bits)

        # default to an IEEE double absent more information:
        #
        # https://en.wikipedia.org/wiki/Double-precision_floating-point_format
        return driver.type_for_floating_point(fp_total_bits=64,
                                              fp_significand_bits=53)

    elif field.field_type == 'boolean':
        return sqlalchemy.sql.sqltypes.BOOLEAN()
    elif field.field_type == 'string':
        if field.constraints and\
           not isinstance(field.constraints, RecordsSchemaFieldStringConstraints):
            raise SyntaxError(f"Incorrect constraint type: {field.constraints}")

        if field.statistics and\
           not isinstance(field.statistics, RecordsSchemaFieldStringStatistics):
            raise SyntaxError(f"Incorrect statistics type: {field.statistics}")

        string_constraints =\
            cast(Optional[RecordsSchemaFieldStringConstraints], field.constraints)
        string_statistics =\
            cast(Optional[RecordsSchemaFieldStringStatistics], field.statistics)
        n = generate_string_length(string_constraints, string_statistics, driver)
        return sqlalchemy.sql.sqltypes.String(n)
    elif field.field_type == 'date':
        return sqlalchemy.sql.sqltypes.DATE()
    elif field.field_type == 'datetime' or field.field_type == 'datetimetz':
        has_tz = field.field_type == 'datetimetz'
        return driver.type_for_date_plus_time(has_tz=has_tz)
    elif field.field_type == 'time' or field.field_type == 'timetz':
        if driver.supports_time_type():
            has_tz = field.field_type == 'timetz'
            return sqlalchemy.sql.sqltypes.TIME(timezone=has_tz)
        else:
            return sqlalchemy.sql.sqltypes.VARCHAR(8)  # HH:MM AM
    else:
        raise NotImplementedError("Teach me how to handle records schema "
                                  f"type {field.field_type}")


def field_to_sqlalchemy_column(field: 'RecordsSchemaField', driver: 'DBDriver') -> Column:
    name = field.name if driver is None else driver.make_column_name_valid(field.name)
    return Column(name,
                  type_=field.to_sqlalchemy_type(driver),
                  nullable=field.constraints is None or not field.constraints.required)



import pathlib
from ...url.resolver import UrlResolver
from sqlalchemy.engine import Engine, Connection
from ...db import DBDriver
from contextlib import contextmanager
from ..records_format import BaseRecordsFormat
from .table import TableRecordsTarget
from .fileobj import FileobjTarget
from .directory_from_url import DirectoryFromUrlRecordsTarget
from .spectrum import SpectrumRecordsTarget
from .data_url import DataUrlTarget
from typing import Callable, Iterator, Optional, Union, Dict, List, IO, TYPE_CHECKING
from ..existing_table_handling import ExistingTableHandling
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa
    from .google_sheets import GoogleSheetsRecordsTarget  # noqa
    from .table import TableRecordsTarget  # noqa
    from .directory_from_url import DirectoryFromUrlRecordsTarget  # noqa


class RecordsTargets(object):
    """
    These methods produce objects representing the target of a records move.

    Example use:

    .. code-block:: python
       records = job_context.records
       db_engine = job_context.get_default_db_engine()
       url = 's3://some-bucket/some-directory/'
       source = records.sources.directory_from_url(url=url)
       target = records.targets.table(schema_name='myschema',
                                      table_name='mytable',
                                      db_engine=db_engine)
       results = records.move(source, target)
    """
    def __init__(self,
                 url_resolver: UrlResolver,
                 db_driver: Callable[[Union[Engine, Connection]], DBDriver]) -> None:
        self.url_resolver = url_resolver
        self.db_driver = db_driver

    @contextmanager
    def directory_from_url(self,
                           output_url: str,
                           records_format:
                           Optional[BaseRecordsFormat]=None) ->\
            Iterator['DirectoryFromUrlRecordsTarget']:
        """
        :param output_url: Location to write the records directory.  Must be a URL format
        understood by the records_mover.url library, and must be a directory URL that ends with a
        '/'.
        :param records_format: Description of the format of the data files to write out.
        """
        from .directory_from_url import DirectoryFromUrlRecordsTarget  # noqa
        yield DirectoryFromUrlRecordsTarget(output_url=output_url,
                                            records_format=records_format,
                                            url_resolver=self.url_resolver)

    @contextmanager
    def table(self,
              schema_name: str,
              table_name: str,
              db_engine: Engine,
              existing_table_handling: ExistingTableHandling=
              ExistingTableHandling.DELETE_AND_OVERWRITE,
              drop_and_recreate_on_load_error: bool=False,
              add_user_perms_for: Optional[Dict[str, List[str]]]=None,
              add_group_perms_for: Optional[Dict[str, List[str]]]=None) -> \
            Iterator['TableRecordsTarget']:
        """
        :param schema_name: Schema name of a table to write data to.

        :param table_name: Table name of a table to write data to.

        :param db_engine: Database engine to write data to.

        :param existing_table_handling: When loading into a database
        table, controls how any existing table found will be handled.

        :param drop_and_recreate_on_load_error: If True, table load errors
        will attempt to be addressed by dropping the target table and
        reloading the incoming data.

        :param add_user_perms_for: If specified, a table's permissions
        will be set for the specified users. Format should be like
        {'all': ['username1', 'username2'], 'select': ['username3', 'username4']}

        :param add_group_perms_for: If specified, a table's permissions
        will be set for the specified group. Format should be like
        {'all': ['group1', 'group2'], 'select': ['group3', 'group4']}
        """
        from .table import TableRecordsTarget  # noqa
        yield TableRecordsTarget(schema_name=schema_name,
                                 table_name=table_name,
                                 db_engine=db_engine,
                                 db_driver=self.db_driver,
                                 existing_table_handling=existing_table_handling,
                                 drop_and_recreate_on_load_error=drop_and_recreate_on_load_error,
                                 add_user_perms_for=add_user_perms_for,
                                 add_group_perms_for=add_group_perms_for)

    @contextmanager
    def google_sheet(self,
                     spreadsheet_id: str,
                     sheet_name: str,
                     google_cloud_creds:
                     'google.auth.credentials.Credentials') ->\
            Iterator['GoogleSheetsRecordsTarget']:
        """
        :param spreadsheet_id: This is the xyz in
        https://docs.google.com/spreadsheets/d/xyz/edit?ts=5be5b383#gid=abc
        :param sheet_name: This is the label of the particular tab within the Google Sheets
         spreadsheet where the data should go.
        :param google_cloud_creds: Instance of google.auth.credentials.Credentials for Google Cloud
                                   Platform access.
        """
        # see the 'gsheets' extras_require option in setup.py - needed for this!
        from .google_sheets import GoogleSheetsRecordsTarget  # noqa
        yield GoogleSheetsRecordsTarget(spreadsheet_id=spreadsheet_id,
                                        sheet_name=sheet_name,
                                        google_cloud_creds=google_cloud_creds)

    @contextmanager
    def fileobj(self,
                output_fileobj: IO[bytes],
                records_format: BaseRecordsFormat) -> Iterator[FileobjTarget]:
        """
        :param output_fileobj: Stream where the file shoud be written to
        :param records_format: Description of the format of the data files.
        """
        yield FileobjTarget(fileobj=output_fileobj, records_format=records_format)

    @contextmanager
    def data_url(self,
                 output_url: str,
                 records_format: Optional[BaseRecordsFormat]=None) -> Iterator[DataUrlTarget]:
        """
        :param output_url: Location of the data file to write.  Must be a URL format understood by
        the records_mover.url library.
        :param records_format: Description of the required format of the data file to write, or
               None for no preference (may be faster depending on the source).
        """
        output_loc = self.url_resolver.file_url(output_url)
        yield DataUrlTarget(output_loc=output_loc,
                            records_format=records_format)

    @contextmanager
    def local_file(self,
                   filename: str,
                   records_format: Optional[BaseRecordsFormat]=None) ->\
            Iterator['DataUrlTarget']:
        """
        :param filename: File path (relative or absolute) of the data file to unload to.
        :param records_format: Description of the required format of the data file to write, or
               None for no preference (may be faster depending on the source).
        """
        url = pathlib.Path(filename).resolve().as_uri()
        with self.data_url(output_url=url, records_format=records_format) as source:
            yield source

    @contextmanager
    def spectrum(self,
                 schema_name: str,
                 table_name: str,
                 db_engine: Engine,
                 spectrum_base_url: Optional[str]=None,
                 spectrum_rdir_url: Optional[str]=None,
                 existing_table_handling: ExistingTableHandling=
                 ExistingTableHandling.TRUNCATE_AND_OVERWRITE) ->\
            Iterator['SpectrumRecordsTarget']:
        """
        :param schema_name: Schema name of a table to write data to.

        :param table_name: Table name of a table to write data to.

        :param db_engine: Database engine to write data to.

        :param spectrum_base_url: Root S3 URL under which a simple
        directory structure will be created for files to be stored, if
        spectrum_rdir_url is not specified.  Note that when using the
        mover CLI, db-facts may be used to provide a default.

        :param spectrum_rdir_url: S3 URL where a records directory
        with files will be stored; otherwise, use db-facts default if
        exists.  If this is not specified, spectrum_base_url must be.

        :param existing_table_handling: When loading into a database
        table, controls how any existing table found will be handled.

        """
        yield SpectrumRecordsTarget(schema_name=schema_name,
                                    table_name=table_name,
                                    db_engine=db_engine,
                                    db_driver=self.db_driver,
                                    url_resolver=self.url_resolver,
                                    spectrum_base_url=spectrum_base_url,
                                    spectrum_rdir_url=spectrum_rdir_url,
                                    existing_table_handling=existing_table_handling)



from sqlalchemy_redshift.commands import CopyCommand
from ...records.records_directory import RecordsDirectory
from ...records.records_format import BaseRecordsFormat, DelimitedRecordsFormat
from ...records.processing_instructions import ProcessingInstructions
import sqlalchemy
from sqlalchemy.schema import Table
import logging
from .records_copy import redshift_copy_options
from ...records.load_plan import RecordsLoadPlan
from ..errors import CredsDoNotSupportS3Import
from typing import Optional, Union, Callable, ContextManager, List
from ...url import BaseDirectoryUrl
from botocore.credentials import Credentials
from ...records.hints import complain_on_unhandled_hints

logger = logging.getLogger(__name__)


class RedshiftLoader:
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                 meta: sqlalchemy.MetaData,
                 temporary_s3_directory_loc: Callable[[], ContextManager[BaseDirectoryUrl]])\
            -> None:
        self.db = db
        self.meta = meta
        self.temporary_s3_directory_loc = temporary_s3_directory_loc

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> Optional[int]:
        if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
            raise NotImplementedError('Teach me how to load '
                                      f'{load_plan.records_format.format_type} format')

        if directory.scheme != 's3':
            with self.temporary_s3_directory_loc() as temp_s3_loc:
                s3_directory = directory.copy_to(temp_s3_loc)
                return self.load(schema=schema,
                                 table=table,
                                 load_plan=load_plan,
                                 directory=s3_directory)

        to = Table(table, self.meta, schema=schema)  # no autoload
        unhandled_hints = set(load_plan.records_format.hints.keys())
        processing_instructions = load_plan.processing_instructions
        redshift_options = redshift_copy_options(unhandled_hints, load_plan.records_format.hints,
                                                 processing_instructions.fail_if_cant_handle_hint,
                                                 processing_instructions.fail_if_row_invalid,
                                                 processing_instructions.max_failure_rows)

        complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                    unhandled_hints, load_plan.records_format.hints)
        # http://sqlalchemy-redshift.readthedocs.io/en/latest/commands.html
        loc = directory.loc
        if not callable(getattr(loc, 'aws_creds', None)):
            raise NotImplementedError('Redshift can only load from an S3 bucket')
        else:
            aws_creds: Optional[Credentials] = directory.loc.aws_creds()  # type: ignore
            if aws_creds is None:
                raise CredsDoNotSupportS3Import('Please provide AWS credentials '
                                                '(run "aws configure")')
            copy = CopyCommand(to=to, data_location=directory.loc.url + '_manifest',
                               access_key_id=aws_creds.access_key,
                               secret_access_key=aws_creds.secret_key,
                               session_token=aws_creds.token, manifest=True,
                               region=directory.loc.region,  # type: ignore
                               empty_as_null=True,
                               **redshift_options)
            logger.info(f"Starting Redshift COPY from {directory}...")
            self.db.execute(copy)
            logger.info("Redshift COPY complete.")
            return None  # redshift doesn't give reliable info on load results

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        try:
            processing_instructions = ProcessingInstructions()
            load_plan = RecordsLoadPlan(records_format=source_records_format,
                                        processing_instructions=processing_instructions)
            if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
                return False
            unhandled_hints = set(load_plan.records_format.hints.keys())
            processing_instructions = load_plan.processing_instructions
            redshift_copy_options(unhandled_hints, load_plan.records_format.hints,
                                  processing_instructions.fail_if_cant_handle_hint,
                                  processing_instructions.fail_if_row_invalid,
                                  processing_instructions.max_failure_rows)
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, load_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return [DelimitedRecordsFormat(variant='bluelabs')]



from sqlalchemy_redshift.commands import UnloadFromSelect
from ...records.records_directory import RecordsDirectory
import sqlalchemy
from sqlalchemy.sql import text
from sqlalchemy.schema import Table
import logging
from records_mover.db.quoting import quote_schema_and_table
from .records_unload import redshift_unload_options
from ...records.unload_plan import RecordsUnloadPlan
from ...records.records_format import (
    BaseRecordsFormat, DelimitedRecordsFormat, ParquetRecordsFormat
)
from typing import Union, Callable, Optional, ContextManager, List
from ...url.base import BaseDirectoryUrl
from botocore.credentials import Credentials
from ..errors import CredsDoNotSupportS3Export
from ...records.hints import complain_on_unhandled_hints
from ..unloader import Unloader


logger = logging.getLogger(__name__)


class RedshiftUnloader(Unloader):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                 table: Callable[[str, str], Table],
                 temporary_s3_directory_loc: Callable[[], ContextManager[BaseDirectoryUrl]],
                 **kwargs) -> None:
        super().__init__(db=db)
        self.table = table
        self.temporary_s3_directory_loc = temporary_s3_directory_loc

    def unload_to_s3_directory(self,
                               schema: str,
                               table: str,
                               unload_plan: RecordsUnloadPlan,
                               directory: RecordsDirectory) -> int:
        unhandled_hints = set()
        if isinstance(unload_plan.records_format, DelimitedRecordsFormat):
            unhandled_hints = set(unload_plan.records_format.hints.keys())
        processing_instructions = unload_plan.processing_instructions
        redshift_options = redshift_unload_options(unhandled_hints,
                                                   unload_plan.records_format,
                                                   processing_instructions.fail_if_cant_handle_hint)
        if isinstance(unload_plan.records_format, DelimitedRecordsFormat):
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, unload_plan.records_format.hints)
        # http://sqlalchemy-redshift.readthedocs.io/en/latest/commands.html
        loc = directory.loc
        if not callable(getattr(loc, 'aws_creds')):
            raise NotImplementedError('Redshift can only load from an S3 bucket')
        else:
            aws_creds: Optional[Credentials] = loc.aws_creds()  # type: ignore
            if aws_creds is None:
                raise CredsDoNotSupportS3Export('Please provide AWS credentials '
                                                '(run "aws configure")')
            select = text(f"SELECT * FROM {quote_schema_and_table(self.db, schema, table)}")
            unload = UnloadFromSelect(select=select,
                                      access_key_id=aws_creds.access_key,
                                      secret_access_key=aws_creds.secret_key,
                                      session_token=aws_creds.token, manifest=True,
                                      unload_location=directory.loc.url, **redshift_options)
            self.db.execute(unload)
            out = self.db.execute(text("SELECT pg_last_unload_count()"))
            rows: Optional[int] = out.scalar()
            assert rows is not None
            logger.info(f"Just unloaded {rows} rows")
            out.close()

            return rows

    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> int:
        logger.info("Starting Redshift unload...")
        if directory.scheme == 's3':
            s3_directory = directory
            return self.unload_to_s3_directory(schema, table, unload_plan, s3_directory)
        else:
            with self.temporary_s3_directory_loc() as temp_s3_loc:
                s3_directory = RecordsDirectory(records_loc=temp_s3_loc)
                out = self.unload_to_s3_directory(schema, table, unload_plan, s3_directory)
                directory.copy_from(temp_s3_loc)
                return out

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        return [DelimitedRecordsFormat(variant='bluelabs'), ParquetRecordsFormat()]

    def can_unload_this_format(self, target_records_format: BaseRecordsFormat) -> bool:
        try:
            unload_plan = RecordsUnloadPlan(records_format=target_records_format)
            unhandled_hints = set()
            if isinstance(unload_plan.records_format, DelimitedRecordsFormat):
                unhandled_hints = set(unload_plan.records_format.hints.keys())
            processing_instructions = unload_plan.processing_instructions
            redshift_unload_options(unhandled_hints,
                                    unload_plan.records_format,
                                    processing_instructions.fail_if_cant_handle_hint)
            if isinstance(unload_plan.records_format, DelimitedRecordsFormat):
                complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                            unhandled_hints,
                                            unload_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False



import boto3
from urllib.parse import urlparse
from airflow.exceptions import AirflowException
from records_mover.records.records import Records
from records_mover.db import db_driver, DBDriver
from records_mover.url.resolver import UrlResolver
from airflow.hooks import BaseHook
from airflow.contrib.hooks.aws_hook import AwsHook
from typing import Optional, Union, List, TYPE_CHECKING
import sqlalchemy
if TYPE_CHECKING:
    from boto3.session import ListObjectsResponseContentType, S3ClientTypeStub  # noqa


class RecordsHook(BaseHook):
    def __init__(self,
                 s3_temp_base_url: Optional[str]=None,
                 aws_conn_id: str='aws_default'):
        self.aws_conn_id = aws_conn_id
        self.__s3_temp_base_url = s3_temp_base_url
        self._boto3_session: Optional[boto3.session.Session] = None

    def _get_boto3_session(self) -> boto3.session.Session:
        if not self._boto3_session:
            self._boto3_session = AwsHook(self.aws_conn_id).get_session()
        return self._boto3_session

    @property
    def _url_resolver(self) -> UrlResolver:
        return UrlResolver(boto3_session=self._get_boto3_session())

    def _db_driver(self, db: Union[sqlalchemy.engine.Engine,
                                   sqlalchemy.engine.Connection]) -> DBDriver:
        s3_temp_base_loc = (self._url_resolver.directory_url(self._s3_temp_base_url)
                            if self._s3_temp_base_url else None)

        return db_driver(db=db, url_resolver=self._url_resolver, s3_temp_base_loc=s3_temp_base_loc)

    @property
    def _s3_temp_base_url(self) -> Optional[str]:
        if self.__s3_temp_base_url and not self.__s3_temp_base_url.endswith('/'):
            raise ValueError("Please provide a directory name - "
                             f"URL should end with '/': {self.__s3_temp_base_url}")
        return self.__s3_temp_base_url

    def validate_and_prepare_target_directory(self, target_url: str,
                                              allow_overwrite: bool=False) -> None:
        parsed_target = urlparse(target_url)
        if parsed_target.scheme != 's3':
            raise AirflowException(
                f"unsupported scheme '{parsed_target.scheme}': {target_url}")
        target_bucket = parsed_target.netloc
        target_key = (parsed_target.path + '/').strip('/')

        s3: 'S3ClientTypeStub' = self._get_boto3_session().client('s3')
        objects: List['ListObjectsResponseContentType'] = s3.list_objects_v2(
            Bucket=target_bucket,
            Prefix=target_key,
        ).get('Contents', [])
        if len(objects) > 0:
            self.log.info("Target URL %s is non-empty (%d keys)", target_url, len(objects))
            if allow_overwrite:
                self.log.warning("Deleting contents of target %s", target_url)
                s3.delete_objects(
                    Bucket=target_bucket,
                    Delete={"Objects": [{"Key": obj["Key"]} for obj in objects]}
                )
            else:
                raise AirflowException('Target URL %s is non-empty', target_url)

    def get_conn(self) -> Records:
        return Records(
            db_driver=self._db_driver,
            url_resolver=self._url_resolver,
        )



import math

INT8_MAX = 127
INT8_MIN = -128
UINT8_MAX = 255
UINT8_MIN = 0
INT16_MAX = 32767
INT16_MIN = -32768
UINT16_MAX = 65535
UINT16_MIN = 0
INT32_MAX = 2147483647
INT32_MIN = -2147483648
UINT32_MAX = 4294967295
UINT32_MIN = 0
INT64_MAX = 9223372036854775807
INT64_MIN = -9223372036854775808
UINT64_MAX = 18446744073709551615
UINT64_MIN = 0
FLOAT16_SIGNIFICAND_BITS = 11
FLOAT32_SIGNIFICAND_BITS = 23
FLOAT64_SIGNIFICAND_BITS = 53
FLOAT80_SIGNIFICAND_BITS = 64


# https://stackoverflow.com/questions/2189800/length-of-an-integer-in-python
def num_digits(n: int) -> int:
    if n > 0:
        digits = int(math.log10(n)) + 1
    elif n == 0:
        digits = 1
    else:
        digits = int(math.log10(-n)) + 2  # +1 if you don't count the '-'
    return digits



from ...utils import quiet_remove
from ...records.hints import cant_handle_hint
from sqlalchemy_redshift.commands import Format
from typing import Dict, Optional, Set
from ...records import RecordsHints

RedshiftCopyOptions = Dict[str, Optional[object]]


def redshift_copy_options(unhandled_hints: Set[str],
                          hints: RecordsHints,
                          fail_if_cant_handle_hint: bool,
                          fail_if_row_invalid: bool,
                          max_failure_rows: Optional[int]) -> RedshiftCopyOptions:
    redshift_options: RedshiftCopyOptions = {}
    redshift_options['compression'] = hints['compression']
    quiet_remove(unhandled_hints, 'compression')
    if hints['dateformat'] is None:
        redshift_options['date_format'] = 'auto'
    else:
        redshift_options['date_format'] = hints['dateformat']
    quiet_remove(unhandled_hints, 'dateformat')
    if hints['encoding'] not in ['UTF8', 'UTF16', 'UTF16LE', 'UTF16BE']:
        cant_handle_hint(fail_if_cant_handle_hint, 'encoding', hints)
        redshift_options['encoding'] = 'UTF8'
    else:
        redshift_options['encoding'] = hints['encoding']
    quiet_remove(unhandled_hints, 'encoding')
    redshift_options['quote'] = hints['quotechar']
    quiet_remove(unhandled_hints, 'quotechar')
    if hints['quoting'] == 'minimal':
        if hints['escape'] is not None:
            cant_handle_hint(fail_if_cant_handle_hint, 'escape', hints)
        if hints['field-delimiter'] != ',':
            cant_handle_hint(fail_if_cant_handle_hint, 'field-delimiter', hints)
        if hints['doublequote'] is not True:
            cant_handle_hint(fail_if_cant_handle_hint, 'doublequote', hints)

        redshift_options['format'] = Format.csv
    else:
        redshift_options['delimiter'] = hints['field-delimiter']
        if hints['escape'] == '\\':
            redshift_options['escape'] = True
        elif hints['escape'] is None:
            redshift_options['escape'] = False
        else:
            cant_handle_hint(fail_if_cant_handle_hint, 'escape', hints)
            redshift_options['escape'] = False
        if hints['quoting'] == 'all':
            redshift_options['remove_quotes'] = True
            if hints['doublequote'] is not False:
                cant_handle_hint(fail_if_cant_handle_hint, 'doublequote', hints)

        elif hints['quoting'] is None:
            redshift_options['remove_quotes'] = False
        else:
            cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
    quiet_remove(unhandled_hints, 'quoting')
    quiet_remove(unhandled_hints, 'escape')
    quiet_remove(unhandled_hints, 'field-delimiter')
    quiet_remove(unhandled_hints, 'doublequote')
    if hints['datetimeformat'] is None:
        redshift_options['time_format'] = 'auto'
    else:
        # After testing, Redshift's date/time parsing doesn't actually
        # support timezone parsing if you give it configuration - as
        # documented below, it doesn't accept a time zone as part of
        # the format string, and in experimentation, it silently drops
        # the offset when data into a timestamptz field if you specify
        # one directly.
        #
        # Its automatic parser seems to be smarter, though, and is
        # likely to handle a variety of formats:
        #
        # https://docs.aws.amazon.com/redshift/latest/dg/automatic-recognition.html
        if hints['datetimeformat'] != hints['datetimeformattz']:
            # The Redshift auto parser seems to take a good handling
            # at our various supported formats, so let's give it a
            # shot if we're not able to specify a specific format due
            # to the Redshift timestamptz limitation:
            #
            # analytics=> create table formattest (test char(32));
            # CREATE TABLE
            # analytics=> insert into formattest values('2018-01-01 12:34:56');
            # INSERT 0 1
            # analytics=> insert into formattest values('01/02/18 15:34');
            # INSERT 0 1
            # analytics=> insert into formattest values('2018-01-02 15:34:12');
            # INSERT 0 1
            # analytics=> insert into formattest values('2018-01-02 10:34 PM');
            # INSERT 0 1
            # analytics=> select test, cast(test as timestamp) as timestamp,
            #             cast(test as date) as date from formattest;
            #
            #                test               |      timestamp      |    date
            # ----------------------------------+---------------------+------------
            #  2018-01-01 12:34:56              | 2018-01-01 12:34:56 | 2018-01-01
            #  01/02/18 15:34                   | 2018-01-02 15:34:00 | 2018-01-02
            #  2018-01-02 15:34:12              | 2018-01-02 15:34:12 | 2018-01-02
            #  2018-01-02 10:34 PM              | 2018-01-02 22:34:00 | 2018-01-02
            # (4 rows)
            #
            # analytics=>
            redshift_options['time_format'] = 'auto'
        else:
            redshift_options['time_format'] = hints['datetimeformat']
    quiet_remove(unhandled_hints, 'datetimeformat')
    quiet_remove(unhandled_hints, 'datetimeformattz')
    # Redshift doesn't support time-only fields, so these will
    # come in as strings regardless.
    quiet_remove(unhandled_hints, 'timeonlyformat')
    if max_failure_rows is not None:
        redshift_options['max_error'] = max_failure_rows
    elif fail_if_row_invalid:
        redshift_options['max_error'] = 0
    else:
        # max allowed value
        redshift_options['max_error'] = 100000
    if hints['record-terminator'] is not None and \
       hints['record-terminator'] != "\n":

        cant_handle_hint(fail_if_cant_handle_hint, 'record-terminator', hints)
    quiet_remove(unhandled_hints, 'record-terminator')

    if hints['header-row']:
        redshift_options['ignore_header'] = 1
    else:
        redshift_options['ignore_header'] = 0
    quiet_remove(unhandled_hints, 'header-row')

    return redshift_options



from typing import Optional

# An arbitrary 4 mb csv I looked at ran around 100,000 lines.
# Assuming we want to limit our memory usage to, say, 400MB of memory,
# let's limit inference to 1,000,000 lines.
DEFAULT_MAX_SAMPLE_SIZE = 1000000


class ProcessingInstructions:
    def __init__(self,
                 fail_if_dont_understand: bool=True,
                 fail_if_cant_handle_hint: bool=True,
                 fail_if_row_invalid: bool=True,
                 max_inference_rows: Optional[int]=DEFAULT_MAX_SAMPLE_SIZE,
                 max_failure_rows: Optional[int]=None) -> None:
        """Directives on how to handle different situations when processing
        records.

        :param fail_if_dont_understand: If True, and a part of the
        RecordsFormat is not understood while processing, then
        immediately fail and raise an exception.  Otherwise, ignore
        the misunderstood instruction (e.g., ignore the hint, assume
        default variant, etc etc)

        :param fail_if_cant_handle_hint: If True, and for whatever
        reason (e.g., limited options in whatever
        library/tool/database is being used) a certain hint can't be
        handled as specified, raise an exception.  Otherwise, ignore
        the hint and use implementation-specific different behavior.

        :param fail_if_row_invalid: If True, and a particular row of
        data in the records file cannot be understood by the library,
        raise an exception.  Otherwise, ignore the row and continue
        and try to load other rows.

        :param max_failure_rows: Sets a tolerance level for number of
        rows of data in the records file that cannot be understood
        by the library that should be ignored. After reaching level,
        raise an exception

        :param max_inference_rows: If the schema is not provided and
        we need it (e.g., we're to load the records into a database
        and there's no existing table), we'll figure it out through
        'type inference' - looking at a bunch of examples of data and
        building a specific schema that can load those rows.  This can
        take some time, so this parameter controls the maximum number
        of rows we'll look at.  Higher values will be more likely to
        result in a schema that can be loaded into, but will take
        longer to load.  If set to None, the entire file will be
        processed.
        """

        self.fail_if_dont_understand = fail_if_dont_understand
        self.fail_if_cant_handle_hint = fail_if_cant_handle_hint
        self.fail_if_row_invalid = fail_if_row_invalid
        self.max_failure_rows = max_failure_rows
        self.max_inference_rows = max_inference_rows



from .base_job_context import BaseJobContext
from .creds.creds_via_lastpass import CredsViaLastPass
from .creds.base_creds import BaseCreds


class IntegrationTestJobContext(BaseJobContext):
    """
    Bare bones job context for use during integration testing that
    doesn't assume the BlueLabs macOS desktop environment.
    """
    @property
    def creds(self) -> BaseCreds:
        return CredsViaLastPass()



from ..url.base import BaseDirectoryUrl
import logging

logger = logging.getLogger(__name__)


class RecordsSchemaSqlFile:
    def __init__(self, records_loc: BaseDirectoryUrl) -> None:
        self.schema_loc = records_loc.file_in_this_directory('_schema')

    def load_schema_sql(self) -> str:
        return self.schema_loc.string_contents()



from typing import Iterable, List, IO, Optional
from pandas.io.parsers import TextFileReader
import chardet
from . import RecordsHints, BootstrappingRecordsHints
from .csv_streamer import stream_csv, python_encoding_from_hint
import io
import logging

logger = logging.getLogger(__name__)


def complain_on_unhandled_hints(fail_if_dont_understand: bool,
                                unhandled_hints: Iterable[str],
                                hints: RecordsHints) -> None:
    unhandled_bindings = [f"{k}={hints[k]}" for k in unhandled_hints]
    unhandled_bindings_str = ", ".join(unhandled_bindings)
    if len(unhandled_bindings) > 0:
        if fail_if_dont_understand:
            err = "Implement these records_format hints or try again with " +\
                f"fail_if_dont_understand=False': {unhandled_bindings_str}"
            raise NotImplementedError(err)
        else:
            logger.warning(f"Did not understand these hints: {unhandled_bindings_str}")


def cant_handle_hint(fail_if_cant_handle_hint: bool, hint_name: str, hints: RecordsHints) -> None:
    if not fail_if_cant_handle_hint:
        logger.warning("Ignoring hint {hint_name} = {hint_value}"
                       .format(hint_name=hint_name,
                               hint_value=repr(hints[hint_name])))
    else:
        raise NotImplementedError(f"Implement hint {hint_name}={repr(hints[hint_name])} " +
                                  "or try again with fail_if_cant_handle_hint=False")


hint_encoding_from_pandas = {
    'utf-8': 'UTF8',
    'utf-16': 'UTF16',
    'utf-16-le': 'UTF16LE',
    'utf-16-be': 'UTF16BE',
}

hint_encoding_from_chardet = {
    'UTF-8-SIG': 'UTF8BOM',
    'UTF-16': 'UTF16',
    'ISO-8859-1': 'LATIN1',
    # For some reason this is lowercase:
    #  https://github.com/chardet/chardet/blob/17218468eb16b7d0068bce7e4d20bac70f0bf555/chardet/utf8prober.py#L51
    'utf-8': 'UTF8',
    # But let's be ready if they change their minds:
    'UTF-8': 'UTF8',
    'Windows-1252': 'CP1252',
}

hint_compression_from_pandas = {
    # https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html
    # https://github.com/bluelabsio/knowledge/
    #    blob/master/Engineering/Architecture/JobDataExchange/output-design.md#hints
    'gzip': 'GZIP',
    'bz2': 'BZIP',
    None: None,
}


def csv_hints_from_reader(reader: TextFileReader) -> RecordsHints:
    header = reader._engine.header
    quotechar = reader._engine.data.dialect.quotechar
    delimiter = reader._engine.data.dialect.delimiter
    escape_char = reader._engine.data.dialect.escapechar
    compression = reader._engine.compression
    encoding = reader._engine.encoding
    doublequote = reader._engine.doublequote

    return {
        'header-row': True if header is not None else False,
        'field-delimiter': delimiter,
        'compression': hint_compression_from_pandas[compression],
        'quotechar': quotechar,
        'doublequote': doublequote,
        'escape': escape_char,
        'encoding': hint_encoding_from_pandas.get(encoding, encoding),
        'dateformat': 'YYYY-MM-DD',
        'timeonlyformat': 'HH12:MI AM',
        'datetimeformat': 'YYYY-MM-DD HH:MI:SS',
        'datetimeformattz': 'YYYY-MM-DD HH:MI:SSOF',
    }


def sniff_hints_from_fileobjs(fileobjs: List[IO[bytes]],
                              initial_hints: BootstrappingRecordsHints) -> RecordsHints:
    if len(fileobjs) != 1:
        # https://app.asana.com/0/53283930106309/1131698268455054
        raise NotImplementedError('Cannot currently sniff hints from mulitple '
                                  'files--please provide hints')
    fileobj = fileobjs[0]
    if not fileobj.seekable():
        raise NotImplementedError('Cannot currently sniff hints from a pure stream--'
                                  'please save file to disk and load from there or '
                                  'provide explicit records format information')
    hints = sniff_hints(fileobj, initial_hints=initial_hints)
    fileobj.seek(0)
    return hints


def infer_newline_format(fileobj: IO[bytes],
                         inferred_hints: RecordsHints,
                         encoding_hint: str) -> None:
    closed = False
    if getattr(fileobj, 'closed', None) is not None:
        closed = fileobj.closed
    if closed or not fileobj.seekable():
        logger.warning("Assuming UNIX newline format, as stream is not rewindable")
        return
    python_encoding = python_encoding_from_hint[encoding_hint]
    original_position = fileobj.tell()
    fileobj.seek(0)
    text_fileobj = io.TextIOWrapper(fileobj, encoding=python_encoding)
    if text_fileobj.newlines is None:  # ...and it almost certainly will be...
        text_fileobj.readline()  # read enough to know newline format
    # https://www.python.org/dev/peps/pep-0278/
    if text_fileobj.newlines is not None:
        inferred_hints['record-terminator'] = str(text_fileobj.newlines)
        logger.info(f"Inferred record terminator as {repr(text_fileobj.newlines)}")
    else:
        logger.warning("Python could not determine newline format of file.")
    text_fileobj.detach()
    fileobj.seek(original_position)


def other_inferred_csv_hints(fileobj: IO[bytes],
                             encoding_hint: str) -> RecordsHints:
    inferred_hints: RecordsHints = {}
    infer_newline_format(fileobj, inferred_hints, encoding_hint)
    return inferred_hints


def sniff_encoding_hint(fileobj: IO[bytes]) -> Optional[str]:
    if getattr(fileobj, 'closed', None) is not None:
        closed = fileobj.closed
    if closed or not fileobj.seekable():
        logger.warning("Could not use chardet to detect encoding, as stream is not rewindable")
        return None
    original_position = fileobj.tell()
    fileobj.seek(0)
    detector = chardet.UniversalDetector()
    while True:
        chunksize = 512
        chunk = fileobj.read(chunksize)
        detector.feed(chunk)
        if detector.done or len(chunk) < chunksize:
            break
    detector.close()
    fileobj.seek(original_position)
    assert detector.result is not None
    if 'encoding' in detector.result:
        chardet_encoding = detector.result['encoding']
        if chardet_encoding in hint_encoding_from_chardet:
            return hint_encoding_from_chardet[chardet_encoding]
        else:
            logger.warning(f"Got unrecognized encoding from chardet sniffing: {detector.result}")
            return None
    else:
        logger.warning(f"Unable to sniff file encoding using chardet: {detector.result}")
        return None


def sniff_hints(fileobj: IO[bytes],
                initial_hints: BootstrappingRecordsHints) -> RecordsHints:
    if 'encoding' not in initial_hints:
        encoding_hint = sniff_encoding_hint(fileobj)
    else:
        encoding_hint = initial_hints['encoding']

    streaming_hints = initial_hints.copy()
    if encoding_hint is not None:
        streaming_hints['encoding'] = encoding_hint
    with stream_csv(fileobj, streaming_hints) as reader:
        # overwrite hints from reader with user-specified values, as
        # the reader isn't smart enough to remember things like which
        # quoting setting it was told to use...
        pandas_inferred_hints = csv_hints_from_reader(reader)
        final_encoding_hint: str = (encoding_hint or  # type: ignore
                                    pandas_inferred_hints['encoding'])
        return {**pandas_inferred_hints,
                'encoding': final_encoding_hint,
                **other_inferred_csv_hints(fileobj, final_encoding_hint),
                **initial_hints}  # type: ignore



import csv
import io
from contextlib import contextmanager
from pandas import read_csv
from pandas.io.parsers import TextFileReader
from typing import Union, IO, Optional, Iterator
from records_mover.records import BootstrappingRecordsHints


pandas_compression_from_hint = {
    'GZIP': 'gzip',
    'BZIP': 'bz2',
    None: None,
}


pandas_quoting_from_hint = {
    'minimal': csv.QUOTE_MINIMAL,
    'all': csv.QUOTE_ALL,
    'nonnumeric': csv.QUOTE_NONNUMERIC,
    None: csv.QUOTE_NONE
}

python_encoding_from_hint = {
    # valid python names: https://docs.python.org/3/library/codecs.html#standard-encodings
    # valid hints names: https://github.com/bluelabsio/knowledge/blob/master/Engineering/
    #    Architecture/JobDataExchange/output-design.md
    'UTF8': 'utf-8',
    'UTF16': 'utf-16',
    'UTF16LE': 'utf-16-le',
    'UTF16BE': 'utf-16-be',
    'LATIN1': 'latin_1',
    'CP1252': 'cp1252',
    'UTF8BOM': 'utf-8-sig',
    # Python will auto-detect UTF-16 with a BOM, but not UTF-8
    'UTF16BOM': 'utf-16',
}


@contextmanager
def stream_csv(filepath_or_buffer: Union[str, IO[bytes]],
               hints: BootstrappingRecordsHints)\
               -> Iterator[TextFileReader]:
    """Returns a context manager that can be used to generate a full or
    partial dataframe from a CSV.  If partial, it will not read the
    entire CSV file into memory."""

    header_row = hints.get('header-row')
    header: Union[str, int, None]
    if header_row is None:
        # we weren't told...
        header = 'infer'
    elif header_row:
        # there's a header row in the file, so tell Pandas that row 0
        # is the headers:
        header = 0
    else:
        # there's no header row in the file
        header = None
    compression_hint: Optional[str] = hints.get('compression')
    encoding_hint = hints.get('encoding', 'UTF8')
    kwargs = {
        'sep': hints.get('field-delimiter', ','),
        'encoding': python_encoding_from_hint.get(encoding_hint, encoding_hint),
        'header': header,
        'compression': pandas_compression_from_hint[compression_hint],
        'escapechar': hints.get('escape'),
        'prefix': 'untitled_',
        'iterator': True,
        'engine': 'python'
    }
    if 'quoting' in hints:
        quoting: Optional[str] = hints['quoting']
        kwargs['quoting'] = pandas_quoting_from_hint[quoting]
    # The streaming code from pandas demands a text stream if we're
    # dealing with an uncompressed CSV file, and a binary stream
    # if we're dealing with compressed file.
    if isinstance(filepath_or_buffer, str):
        # Send a local filename...
        out = None
        try:
            out = read_csv(filepath_or_buffer, **kwargs)
            yield out
        finally:
            if out is not None:
                out.close()
    elif compression_hint is None:
        hint_encoding: str = hints.get('encoding')  # type: ignore
        encoding = python_encoding_from_hint.get(hint_encoding, hint_encoding)
        text_fileobj = io.TextIOWrapper(filepath_or_buffer, encoding=encoding)
        out = None
        try:
            out = read_csv(text_fileobj, **kwargs)
            yield out
        finally:
            if out is not None:
                out.close()
            text_fileobj.detach()
    else:
        # Pass as a binary file
        out = None
        try:
            out = read_csv(filepath_or_buffer, **kwargs)
            yield out
        finally:
            if out is not None:
                out.close()



import logging
from .s3_base_url import S3BaseUrl
from ..base import BaseDirectoryUrl, BaseFileUrl
from typing import IO, List, Optional
import threading
from time import sleep
from s3_concat import S3Concat
from smart_open.s3 import open as s3_open


logger = logging.getLogger(__name__)


class S3FileUrl(S3BaseUrl, BaseFileUrl):
    def __str__(self) -> str:
        return self.url

    def filename(self) -> str:
        return self.url[self.url.rfind("/")+1:]

    def directory_in_this_directory(self, directory_name: str) -> BaseDirectoryUrl:
        directory = self.containing_directory()
        return self._directory(f"{directory.url}{directory_name}/")

    def file_in_this_directory(self, filename: str) -> 'S3FileUrl':
        return self._file(self.containing_directory().url + filename)

    def upload_fileobj(self, fileobj: IO[bytes], mode: str = 'wb') -> int:
        # https://github.com/boto/boto3/blob/b2affa81c9b55ebcb9cb3af6e928f4f5acf22cb9/docs/source/guide/s3-uploading-files.rst
        if mode != 'wb':
            # use the single-threaded method that handles all modes
            return super().upload_fileobj(fileobj, mode=mode)

        class _S3ProgressTracker:
            def __init__(self):
                self.length = 0
                self._lock = threading.Lock()

            def __call__(self, bytes_amount: int) -> None:
                with self._lock:
                    self.length += bytes_amount

        callback = _S3ProgressTracker()

        self.s3_client.upload_fileobj(Fileobj=fileobj,
                                      Bucket=self.bucket,
                                      Key=self.key,
                                      Callback=callback)

        return callback.length

    def open(self, mode: str = "rb") -> IO[bytes]:
        try:
            return s3_open(bucket_id=self.bucket,
                           key_id=self.key,
                           mode=mode,
                           session=self._boto3_session)
        except ValueError as e:
            # Example: ValueError: 'b0KD9AkG7XA/_manifest' does not
            #  exist in the bucket 'vince-scratch', or is forbidden
            #  for access
            if 'does not exist in the bucket' in str(e):
                raise FileNotFoundError(f"{self} not found")
            else:
                raise e

    def download_fileobj(self, fileobj: IO[bytes]) -> None:
        self.s3_client.download_fileobj(Fileobj=fileobj, Bucket=self.bucket, Key=self.key)

    def store_string(self, contents: str) -> None:
        self.s3_resource.Object(self.bucket, self.key).put(Body=contents)

    def rename_to(self, new: 'BaseFileUrl') -> 'S3FileUrl':
        if not isinstance(new, S3FileUrl):
            raise TypeError(f'Can only rename to same type, not {new}')
        # https://stackoverflow.com/questions/32501995/boto3-s3-renaming-an-object-using-copy-object
        copy_source = {'Bucket': self.bucket, 'Key': self.key}
        self.s3_resource.Object(new.bucket, new.key).copy_from(CopySource=copy_source)
        self.s3_resource.Object(self.bucket, self.key).delete()
        logger.info("Renamed {old_url} to {new_url}".format(old_url=self.url, new_url=new.url))
        return new

    def delete(self) -> None:
        self.s3_resource.Object(self.bucket, self.key).delete()

    def wait_to_exist(self) -> None:
        while True:
            try:
                with self.open():
                    return
            except FileNotFoundError:
                logger.info(f"Waiting for {self.url} to appear...")
                fifty_milliseconds = 0.05
                sleep(fifty_milliseconds)

    def size(self) -> int:
        response = self.s3_client.head_object(Bucket=self.bucket, Key=self.key)
        return response['ContentLength']

    def concatenate_from(self, other_locs: List['BaseFileUrl']) -> Optional[int]:
        if not all([isinstance(loc, S3FileUrl) and loc.bucket == self.bucket
                    for loc in other_locs]):
            logger.warning("Concatenating data locally - this may be slow for large data sets")
            return super().concatenate_from(other_locs)
        job = S3Concat(self.bucket,
                       self.key,
                       session=self._boto3_session,
                       # We want one file the end--S3Concat's other
                       # job in life is concatting small log files
                       # into larger ones, where a minimum file size
                       # is a hint for when to stop combining files
                       # together.
                       min_file_size=None)
        for loc in other_locs:
            assert isinstance(loc, S3FileUrl)  # keep mypy happy
            # Add files, can call multiple times to add files from other directories
            job.add_file(loc.key)

        out = job.concat()
        assert len(out) == 1  # with above arg, this should provide only a single output file

        return None



import csv
from ...utils import quiet_remove
from ..hints import cant_handle_hint
from ..processing_instructions import ProcessingInstructions
from ..records_format import DelimitedRecordsFormat
import logging
from typing import Set, Dict


logger = logging.getLogger(__name__)


def pandas_to_csv_options(records_format: DelimitedRecordsFormat,
                          unhandled_hints: Set[str],
                          processing_instructions: ProcessingInstructions) -> Dict[str, object]:
    # https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_csv.html
    hints = records_format.hints

    fail_if_cant_handle_hint = processing_instructions.fail_if_cant_handle_hint

    pandas_options: Dict[str, object] = {}

    pandas_options['encoding'] = hints['encoding']
    quiet_remove(unhandled_hints, 'encoding')

    if hints['compression'] is None:
        # hints['compression']=None will output an uncompressed csv,
        # which is the pandas default.
        pass
    elif hints['compression'] == 'GZIP':
        pandas_options['compression'] = 'gzip'
    elif hints['compression'] == 'BZIP':
        pandas_options['compression'] = 'bz2'
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)
    quiet_remove(unhandled_hints, 'compression')

    if hints['quoting'] is None:
        pandas_options['quoting'] = csv.QUOTE_NONE
    elif hints['quoting'] == 'all':
        pandas_options['quoting'] = csv.QUOTE_ALL
    elif hints['quoting'] == 'minimal':
        pandas_options['quoting'] = csv.QUOTE_MINIMAL
    elif hints['quoting'] == 'nonnumeric':
        pandas_options['quoting'] = csv.QUOTE_NONNUMERIC
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
    quiet_remove(unhandled_hints, 'quoting')

    pandas_options['doublequote'] = hints['doublequote']
    quiet_remove(unhandled_hints, 'doublequote')
    pandas_options['quotechar'] = hints['quotechar']
    quiet_remove(unhandled_hints, 'quotechar')

    if hints['escape'] is None:
        pass
    else:
        pandas_options['escapechar'] = hints['escape']
    quiet_remove(unhandled_hints, 'escape')

    pandas_options['header'] = hints['header-row']
    quiet_remove(unhandled_hints, 'header-row')

    if hints['dateformat'] is None:
        if hints['datetimeformattz'] == hints['datetimeformat']:
            # BigQuery requires that timezone offsets have a colon;
            # Python (and thus Pandas) doesn't support adding the
            # colon with strftime.  However, we can specify things
            # without a timezone delimiter just fine.
            #
            # Unfortunately Python/Pandas will drop the timezone info
            # instead of converting the timestamp to UTC.  This
            # corrupts the time, as BigQuery assumes what it gets in
            # is UTC format.  Boo.
            #
            # $ python3
            # >>> import pytz
            # >>> us_eastern = pytz.timezone('US/Eastern')
            # >>> import datetime
            # >>> us_eastern.localize(datetime.datetime(2000, 1, 2, 12, 34, 56, 789012))
            #        .strftime('%Y-%m-%d %H:%M:%S.%f')
            # '2000-01-02 12:34:56.789012'
            # >>>
            #
            # https://app.asana.com/0/1128138765527694/1159958019131681
            pandas_options['date_format'] = '%Y-%m-%d %H:%M:%S.%f'
        else:
            pandas_options['date_format'] = '%Y-%m-%d %H:%M:%S.%f%z'
    elif hints['dateformat'] == 'YYYY-MM-DD':
        if hints['datetimeformattz'] == hints['datetimeformat']:
            pandas_options['date_format'] = '%Y-%m-%d %H:%M:%S.%f'
        else:
            pandas_options['date_format'] = '%Y-%m-%d %H:%M:%S.%f%z'
    elif hints['dateformat'] == 'MM-DD-YYYY':
        if hints['datetimeformattz'] == hints['datetimeformat']:
            pandas_options['date_format'] = '%m-%d-%Y %H:%M:%S.%f'
        else:
            pandas_options['date_format'] = '%m-%d-%Y %H:%M:%S.%f%z'
    elif hints['dateformat'] == 'DD-MM-YYYY':
        if hints['datetimeformattz'] == hints['datetimeformat']:
            pandas_options['date_format'] = '%d-%m-%Y %H:%M:%S.%f'
        else:
            pandas_options['date_format'] = '%d-%m-%Y %H:%M:%S.%f%z'
    elif hints['dateformat'] == 'MM/DD/YY':
        if hints['datetimeformattz'] == hints['datetimeformat']:
            pandas_options['date_format'] = '%m/%d/%y %H:%M:%S.%f'
        else:
            pandas_options['date_format'] = '%m/%d/%y %H:%M:%S.%f%z'
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'dateformat', hints)
    quiet_remove(unhandled_hints, 'dateformat')

    # pandas can't seem to export a date and time together :(
    #
    # might be nice someday to only emit the errors if the actual data
    # being moved is affected by whatever limitation...
    if (hints['datetimeformattz'] not in(f"{hints.get('dateformat', 'YYYY-MM-DD')} HH24:MI:SSOF",
                                         f"{hints.get('dateformat', 'YYYY-MM-DD')} HH:MI:SSOF",
                                         f"{hints.get('dateformat', 'YYYY-MM-DD')} HH24:MI:SS",
                                         f"{hints.get('dateformat', 'YYYY-MM-DD')} HH:MI:SS",
                                         f"{hints.get('dateformat', 'YYYY-MM-DD')} HH:MIOF",
                                         f"{hints.get('dateformat', 'YYYY-MM-DD')} HH:MI",
                                         f"{hints.get('dateformat', 'YYYY-MM-DD')} HH24:MIOF",
                                         f"{hints.get('dateformat', 'YYYY-MM-DD')} HH24:MI")):
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformattz', hints)
    quiet_remove(unhandled_hints, 'datetimeformattz')

    valid_datetimeformat = [
            f"{hints.get('dateformat', 'YYYY-MM-DD')} HH24:MI:SS",
            f"{hints.get('dateformat', 'YYYY-MM-DD')} HH:MI:SS",
            f"{hints.get('dateformat', 'YYYY-MM-DD')} HH24:MI",
            f"{hints.get('dateformat', 'YYYY-MM-DD')} HH:MI",
    ]
    if (hints['datetimeformat'] not in valid_datetimeformat):
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', hints)
    quiet_remove(unhandled_hints, 'datetimeformat')

    if hints['timeonlyformat'] != 'HH24:MI:SS':
        cant_handle_hint(fail_if_cant_handle_hint, 'timeonlyformat', hints)
    quiet_remove(unhandled_hints, 'timeonlyformat')

    pandas_options['sep'] = hints['field-delimiter']
    quiet_remove(unhandled_hints, 'field-delimiter')

    pandas_options['line_terminator'] = hints['record-terminator']
    quiet_remove(unhandled_hints, 'record-terminator')

    return pandas_options



from pandas import Series, isnull, DataFrame
from math import ceil
import json
from typing import List


def compress_sparse_dataframe_columns(df: DataFrame,
                                      min_required_populated_ratio: float=0.25,
                                      mandatory: List[str]=[]):
    """Compresses a sparse dataframe by throwing sparse columns into a
    single column containing a dict representing the sparse values.

    min_required_populated_ratio: What ratio of rows must provide this
                                  data element for us to grace it with
                                  a column automatically.

    mandatory: Which columns should be provided always, even if
               there's almost no data for it.

    df:
       a    b    c    d    e    f
    0  0  NaN  NaN  NaN  NaN  NaN
    1  1  1.0  NaN  NaN  NaN  NaN
    2  2  NaN  2.0  NaN  NaN  NaN
    3  3  NaN  NaN  3.0  NaN  NaN
    4  4  NaN  NaN  NaN  4.0  NaN
    5  5  NaN  NaN  NaN  NaN  5.0

    compressed_df:
       a  compressed
    0  0          {}
    1  1  {'b': 1.0}
    2  2  {'c': 2.0}
    3  3  {'d': 3.0}
    4  4  {'e': 4.0}
    5  5  {'f': 5.0}


    After saving dataframe to Redshift:

    analytics=> select * from vbroz.compressed_test
                where json_extract_path_text(compressed, 'c') <> '';
     a | compressed
    ---+------------
     2 | {"c": 2.0}
    (1 row)

    """
    min_rows_to_save = ceil(min_required_populated_ratio * len(df.index))
    core_df = df.dropna(axis=1, how='all', thresh=min_rows_to_save)
    save_these_columns = set(mandatory) - set(core_df)
    save_these_series = {col: df[col] for col in save_these_columns}
    core_df = core_df.assign(**save_these_series)

    kept_columns = set(core_df)
    compressed_df = df.drop(kept_columns, 1)
    wordy_compressed_dicts = compressed_df.to_dict(orient='records')

    def remove_null_keys(d):
        return {k: v for k, v in d.items() if not isnull(v)}

    compressed_dicts = list(map(remove_null_keys, wordy_compressed_dicts))
    compressed_json = list(map(lambda d: json.dumps(d), compressed_dicts))

    return core_df.assign(compressed=Series(compressed_json,
                                            index=core_df.index))



from contextlib import contextmanager
from sqlalchemy.schema import CreateTable
from ..records.records_format import BaseRecordsFormat, DelimitedRecordsFormat
from tempfile import TemporaryDirectory
from ..url.filesystem import FilesystemDirectoryUrl
import logging
import sqlalchemy
from sqlalchemy import MetaData
from sqlalchemy.schema import Table
from ..url.base import BaseDirectoryUrl
from ..records.unload_plan import RecordsUnloadPlan
from ..records.load_plan import RecordsLoadPlan
from ..records.types import RecordsFormatType
from ..records.records_directory import RecordsDirectory
from records_mover.db.quoting import quote_group_name, quote_user_name, quote_schema_and_table
from records_mover.records import RecordsSchema
from typing import Union, Dict, List, IO, Tuple, Iterator, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from typing_extensions import Literal  # noqa

logger = logging.getLogger(__name__)


class DBDriver:
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine,
                           sqlalchemy.engine.Connection], **kwargs) -> None:
        self.db = db
        self.db_engine = db.engine
        self.meta = MetaData()

    def best_scheme_to_load_from(self) -> str:
        return 'file'

    def has_table(self, schema: str, table: str) -> bool:
        return self.db.dialect.has_table(self.db, table_name=table, schema=schema)

    def table(self,
              schema: str,
              table: str) -> Table:
        return Table(table, self.meta, schema=schema, autoload=True, autoload_with=self.db_engine)

    def schema_sql(self,
                   schema: str,
                   table: str) -> str:
        """Generate DDL which will recreate the specified table and return it
        as a string.  Returns None if database or permissions don't
        support operation.

        """
        # http://docs.sqlalchemy.org/en/latest/core/reflection.html
        table_obj = self.table(schema, table)
        return str(CreateTable(table_obj, bind=self.db))

    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> int:
        """Writes table specified to the RecordsDirectory instance named 'directory'
        per the UnloadPlan named 'unload_plan'.  Guarantees a manifest
        file named 'manifest' is written to the target directory pointing
        to the target records.  After ensuring other records-related
        metadata is provided, you can call directory.finalize_manifest()
        to move this to the final location of _manifest, signaling to
        readers that the Records directory is finalized.

        Returns number of rows unloaded.
        """
        raise NotImplementedError(f"unload not implemented for this database type")

    def can_unload_this_format(self, target_records_format: BaseRecordsFormat) -> bool:
        """Return true if the specified format is compatible with the unload()
        method"""
        return False

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        """Candidates to look through when negotiating a common records format
        with a records target"""
        return []

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> Optional[int]:
        """Loads the data from the data specified to the RecordsDirectory
        instance named 'directory'.  Guarantees a manifest file named
        'manifest' is written to the target directory pointing to the
        target records.

        Returns number of rows loaded (if database provides that
        info).
        """
        raise NotImplementedError(f"load not implemented for this database type")

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        """Return true if the specified format is compatible with the load()
        method"""
        return False

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        """Candidates to look through when negotiating a common records format
        with a records source.  Will be looked through in order, so
        the better formats (higher fidelity and/or more efficient to
        process) should appear first in the list.
        """
        return []

    def best_records_format_variant(self,
                                    records_format_type: RecordsFormatType) -> \
            Optional[str]:
        """Return the most suitable records format delimited variant for
        loading and unloading.  This is generally the format that the
        database unloading and loading natively, which won't require
        translation before loading or after unloading.
        """
        if records_format_type == 'delimited':
            return 'bluelabs'
        else:
            return None

    def best_records_format(self) -> BaseRecordsFormat:
        variant = self.best_records_format_variant('delimited')
        assert variant is not None  # always provided for 'delimited'
        return DelimitedRecordsFormat(variant=variant)

    def can_load_from_fileobjs(self):
        return False

    def load_from_fileobj(self, schema: str, table: str,
                          load_plan: RecordsLoadPlan, fileobj: IO[bytes]) -> Optional[int]:
        """Loads the data from the file stream provided.
        """
        raise NotImplementedError(f"load_from_fileobj not implemented for this database type")

    def varchar_length_is_in_chars(self) -> bool:
        """True if the 'n' in VARCHAR(n) is represented in natural language
        characters, rather than in post-encoding bytes.  This varies by database -
        override it to control"""
        return False

    def set_grant_permissions_for_groups(self, schema_name: str, table: str,
                                         groups: Dict[str, List[str]],
                                         db: Union[sqlalchemy.engine.Engine,
                                                   sqlalchemy.engine.Connection]) -> None:
        schema_and_table: str = quote_schema_and_table(self.db.engine, schema_name, table)
        for perm_type in groups:
            groups_list = groups[perm_type]
            for group in groups_list:
                group_name: str = quote_group_name(self.db.engine, group)
                if not perm_type.isalpha():
                    raise TypeError("Please make sure your permission types"
                                    " are an acceptable value.")
                perms_sql = f'GRANT {perm_type} ON TABLE {schema_and_table} TO {group_name}'
                db.execute(perms_sql)

    def set_grant_permissions_for_users(self, schema_name: str, table: str,
                                        users: Dict[str, List[str]],
                                        db: Union[sqlalchemy.engine.Engine,
                                                  sqlalchemy.engine.Connection]) -> None:
        schema_and_table: str = quote_schema_and_table(self.db.engine, schema_name, table)
        for perm_type in users:
            user_list = users[perm_type]
            for user in user_list:
                user_name: str = quote_user_name(self.db.engine, user)
                if not perm_type.isalpha():
                    raise TypeError("Please make sure your permission types"
                                    " are an acceptable value.")
                perms_sql = f'GRANT {perm_type} ON TABLE {schema_and_table} TO {user_name}'
                db.execute(perms_sql)

    def load_failure_exception(self):
        return sqlalchemy.exc.InternalError

    @contextmanager
    def temporary_loadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        with TemporaryDirectory(prefix='temporary_loadable_directory_loc') as dirname:
            yield FilesystemDirectoryUrl(dirname)

    def supports_time_type(self):
        return True

    def type_for_date_plus_time(self, has_tz: bool=False) -> sqlalchemy.sql.sqltypes.DateTime:
        """Different DB vendors have different names for a date, a time, and
        an optional timezone"""
        return sqlalchemy.sql.sqltypes.DateTime(timezone=has_tz)

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit data matching the given
        min and max integer values"""
        logger.warning("Using default integer type")
        return sqlalchemy.sql.sqltypes.Integer()

    def type_for_fixed_point(self,
                             precision: int,
                             scale: int) -> sqlalchemy.sql.sqltypes.Numeric:
        """Find correct decimal column type to fit data from the given fixed point type"""
        return sqlalchemy.sql.sqltypes.Numeric(precision=precision,
                                               scale=scale)

    def type_for_floating_point(self,
                                fp_total_bits: int,
                                fp_significand_bits: int) -> sqlalchemy.sql.sqltypes.Numeric:
        """Find correct decimal column type to fit data from the given floating point type"""

        # SQL spec (at least in the publically available pre-release
        # 1992 version) declares the meaning of the precision in type
        # "FLOAT(precision)" to be defined by the database vendor:
        #
        # 49)Subclause 6.1, "<data type>": The binary precision of a
        # data type defined as FLOAT for each value specified by
        # <precision> is implementation-defined.
        #
        # http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
        #
        # That said, at least Postgres, MySQL and ORacle agree that
        # the 'precision' in their cases mean the significand_bits of
        # either an IEEE or Intel-flavored float , so let's default to
        # that:
        #
        # https://dev.mysql.com/doc/refman/8.0/en/floating-point-types.html#targetText=MySQL%20permits%20a%20nonstandard%20syntax,look%20like%20%2D999.9999%20when%20displayed.
        # https://docs.oracle.com/javadb/10.8.3.0/ref/rrefsqlj27281.html
        # https://www.postgresql.org/docs/10/datatype-numeric.html
        return sqlalchemy.sql.sqltypes.Float(precision=fp_significand_bits)

    def make_column_name_valid(self, colname: str) -> str:
        return colname

    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        """returns the integer limits (min and max value as a tuple) of the
        given column type for this database type.  These are
        represented as a Python int, which is of arbitrary length.

        Since SQLAlchemy doesn't track this detail, if details of this
        database type aren't known in the DBDriver hierarchy, None
        will be returned.
        """
        logger.warning(f"{type(self)} does not know how to gather limits of {type(type_)}")
        return None

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        """returns the floating point representation (total floating point bits minus padding
        and significand bits as a tuple) of the given column type for
        this database type.

        Since SQLAlchemy doesn't track this detail, if details of this
        database type aren't known in the DBDriver hierarchy, None
        will be returned.
        """
        logger.warning(f"{type(self)} does not know how to gather limits of {type(type_)}")
        return None

    def fixed_point_constraints(self,
                                type_: sqlalchemy.types.Numeric) ->\
            Optional[Tuple[int, int]]:
        """returns the fixed-point representation - total number of digits
        (precision) and number of those digits to the right of the
        decimal point (scale) - as a tuple for the given column type
        for this database type.
        """
        if type_.precision is None or type_.scale is None:
            return None
        return (type_.precision, type_.scale)

    def tweak_records_schema_for_load(self,
                                      records_schema: RecordsSchema,
                                      records_format: BaseRecordsFormat) -> RecordsSchema:
        return records_schema



from typing import List, TypeVar, MutableSet, Union
import warnings
import functools


T = TypeVar('T')


def quiet_remove(coll: Union[List[T], MutableSet[T]], element: T) -> None:
    try:
        coll.remove(element)
    except KeyError:
        pass


# https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)
    return new_func


class BetaWarning(Warning):
    pass


def beta(func):
    """This is a decorator which can be used to mark functions as beta
    (interface may change in future). It will result in a warning
    being emitted when the function is used.
    """
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', BetaWarning)  # turn off filter
        warnings.warn("Call to beta function {} - "
                      "interface may change in the future!.".format(func.__name__),
                      category=BetaWarning,
                      stacklevel=2)
        warnings.simplefilter('default', BetaWarning)  # reset filter
        return func(*args, **kwargs)
    return new_func



from ..creds.base_creds import BaseCreds
from ..creds.creds_via_airflow import CredsViaAirflow
from ..base_job_context import BaseJobContext, JsonSchema
from typing import Optional


class AirflowJobContext(BaseJobContext):
    __creds_via_airflow_value: Optional[CredsViaAirflow] = None

    def __init__(self,
                 name: str,
                 default_db_creds_name: Optional[str],
                 default_aws_creds_name: Optional[str],
                 config_json_schema: Optional[JsonSchema],
                 scratch_s3_url: Optional[str] = None) -> None:
        if default_aws_creds_name is None:
            default_aws_creds_name = 'aws_default'
        super().__init__(name=name,
                         default_db_creds_name=default_db_creds_name,
                         default_aws_creds_name=default_aws_creds_name,
                         scratch_s3_url=scratch_s3_url,
                         config_json_schema=config_json_schema)
        self.__request_config = {}

    @property
    def creds(self) -> BaseCreds:
        return self.creds_via_airflow

    @property
    def creds_via_airflow(self) -> CredsViaAirflow:
        if self.__creds_via_airflow_value is None:
            self.__creds_via_airflow_value = CredsViaAirflow()
        return self.__creds_via_airflow_value



from subprocess import check_output
from records_mover.db.db_type import canonicalize_db_type, db_protocol
from ..utils import deprecated
from db_facts.db_facts_types import DBFacts


# This function was written under Python 2.7 and never got fixed to be
# Python 3 compatible, so it kept returning a byte array when it
# should have been returning a string.  It also never chomped the last
# newline off of the command line output, so the API to it was never
# great.  Meanwhile, clients of it started to use it and then expect
# the results to be byte arrays, so just changing behavior isn't a
# great option regardless.
#
# So, let's deprecate it until the next major version here.
#
# Please switch code over to use lpass_field instead.
@deprecated
def lpass_show(name: str, field: str) -> bytes:
    if field == 'notes':
        field_arg = '--notes'
    elif field == 'username':
        field_arg = '--username'
    elif field == 'password':
        field_arg = '--password'
    elif field == 'url':
        field_arg = '--url'
    else:
        field_arg = '--field=' + field
    return check_output(['lpass',
                         'show',
                         field_arg,
                         name])


def lpass_field(name: str, field: str) -> str:
    if field == 'notes':
        field_arg = '--notes'
    elif field == 'username':
        field_arg = '--username'
    elif field == 'password':
        field_arg = '--password'
    elif field == 'url':
        field_arg = '--url'
    else:
        field_arg = '--field=' + field
    raw_output = check_output(['lpass',
                               'show',
                               field_arg,
                               name])
    return raw_output.decode('utf-8').rstrip('\n')


def db_facts_from_lpass(lpass_entry_name: str) -> DBFacts:
    user = lpass_field(lpass_entry_name, 'username')
    password = lpass_field(lpass_entry_name, 'password')
    host = lpass_field(lpass_entry_name, 'Hostname')
    port = int(lpass_field(lpass_entry_name, 'Port'))
    raw_db_type = lpass_field(lpass_entry_name, 'Type')
    db_type = canonicalize_db_type(raw_db_type)
    dbname = lpass_field(lpass_entry_name, 'Database')

    return {'password': password,
            'host': host,
            'user': user,
            'type': db_type,
            'protocol': db_protocol(db_type),
            'port': port,
            'database': dbname}



from .base_job_context import BaseJobContext
from .creds.creds_via_env import CredsViaEnv
from .creds.base_creds import BaseCreds


class EnvJobContext(BaseJobContext):
    """
    Job context which obtains secrets from env variables.
    """
    @property
    def creds(self) -> BaseCreds:
        return CredsViaEnv()



import io
from shutil import copyfileobj
from tempfile import NamedTemporaryFile
from .base import SupportsMoveFromDataframes
from ..results import MoveResult
from ..processing_instructions import ProcessingInstructions
from ..pandas import pandas_to_csv_options
from ..records_format import BaseRecordsFormat, DelimitedRecordsFormat
from ..hints import complain_on_unhandled_hints
import logging
from ..sources.dataframes import DataframesRecordsSource
from typing import IO, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from pandas import DataFrame  # noqa

logger = logging.getLogger(__name__)


class FileobjTarget(SupportsMoveFromDataframes):
    def __init__(self,
                 fileobj: IO[bytes],
                 records_format: BaseRecordsFormat) -> None:
        self.fileobj = fileobj
        self.records_format = records_format

    def move_from_dataframes_source(self,
                                    dfs_source: DataframesRecordsSource,
                                    processing_instructions:
                                    ProcessingInstructions) -> MoveResult:
        if not isinstance(self.records_format, DelimitedRecordsFormat):
            raise NotImplementedError("Teach me to export from dataframe to "
                                      f"{self.records_format.format_type}")
        unhandled_hints = set(self.records_format.hints.keys())
        options = pandas_to_csv_options(self.records_format,
                                        unhandled_hints,
                                        processing_instructions)
        complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                    unhandled_hints, self.records_format.hints)
        logger.info(f"Writing CSV file to {self.fileobj} with options {options}...")
        encoding: str = self.records_format.hints['encoding']  # type: ignore

        def write_dfs(path_or_buf: Union[str, IO[str]]) -> int:
            first_row = True
            move_count = 0
            for df in dfs_source.dfs:
                logger.info("Appending from dataframe...")

                # Include the header at most once in the file
                include_header_row = options['header'] and first_row
                first_row = False
                options['header'] = include_header_row
                df.to_csv(path_or_buf=path_or_buf,
                          mode="a",
                          index=dfs_source.include_index,
                          **options)
                move_count += len(df.index)
            return move_count

        # Pandas' df.to_csv won't write a compressed file to a stream
        # (bleh).  Instead, if compression is set, write to a temp
        # file first.
        if self.records_format.hints['compression'] is None:
            text_fileobj = io.TextIOWrapper(self.fileobj, encoding=encoding)
            move_count = write_dfs(text_fileobj)
            text_fileobj.detach()
        else:
            with NamedTemporaryFile(prefix='mover_fileobj_target') as output_file:
                move_count = write_dfs(output_file.name)
                with open(output_file.name, "rb") as output_fileobj:
                    copyfileobj(output_fileobj, self.fileobj)

        logger.info('CSV file written')
        return MoveResult(output_urls=None, move_count=move_count)



from contextlib import contextmanager
from .base import (SupportsMoveToRecordsDirectory,
                   SupportsToDataframesSource)
from ..records_directory import RecordsDirectory
from ...utils.concat_files import ConcatFiles
import io
from ..results import MoveResult
from ..records_format import BaseRecordsFormat, DelimitedRecordsFormat
import pandas as pd
from ..hints import sniff_hints_from_fileobjs
from .. import BootstrappingRecordsHints
from ..processing_instructions import ProcessingInstructions
from ...records.hints import complain_on_unhandled_hints
from ..csv_streamer import python_encoding_from_hint
from ..schema import RecordsSchema
from ..pandas import pandas_read_csv_options
import logging
from typing import Mapping, IO, Optional, Iterator, List, Any, TYPE_CHECKING
if TYPE_CHECKING:
    from .dataframes import DataframesRecordsSource  # noqa


logger = logging.getLogger(__name__)


class FileobjsSource(SupportsMoveToRecordsDirectory,
                     SupportsToDataframesSource):
    def __init__(self,
                 target_names_to_input_fileobjs: Mapping[str, IO[bytes]],
                 records_format: BaseRecordsFormat,
                 records_schema: RecordsSchema) -> None:
        self.target_names_to_input_fileobjs = target_names_to_input_fileobjs
        self.records_format = records_format
        self.records_schema = records_schema

    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        return [self.records_format]

    def can_move_to_this_format(self,
                                target_records_format: BaseRecordsFormat) -> bool:
        return self.records_format == target_records_format

    @staticmethod
    @contextmanager
    def infer_if_needed(target_names_to_input_fileobjs: Mapping[str, IO[bytes]],
                        processing_instructions: ProcessingInstructions,
                        records_format: Optional[BaseRecordsFormat],
                        records_schema: Optional[RecordsSchema],
                        initial_hints: Optional[BootstrappingRecordsHints]) ->\
            Iterator['FileobjsSource']:
        try:
            if records_format is None:
                if initial_hints is None:
                    initial_hints = {}
                logger.info(f"Determining records format with initial_hints={initial_hints}")
                inferred_hints =\
                    sniff_hints_from_fileobjs(list(target_names_to_input_fileobjs.values()),
                                              initial_hints=initial_hints)
                # 'csv' isn't the most precise variant or fastest
                # variant to read, but given it's the default for Excel
                # and Google Sheets, it's the most common on import.  So,
                # if the specific variant isn't specified by the user,
                # let's assume 'csv'.
                records_format = DelimitedRecordsFormat(variant='csv',
                                                        hints=inferred_hints)
            if records_schema is None:
                records_schema =\
                    RecordsSchema.from_fileobjs(list(target_names_to_input_fileobjs.values()),
                                                records_format=records_format,
                                                processing_instructions=processing_instructions)

            yield FileobjsSource(target_names_to_input_fileobjs=target_names_to_input_fileobjs,
                                 records_format=records_format,
                                 records_schema=records_schema)
        except UnicodeDecodeError:
            if isinstance(records_format, DelimitedRecordsFormat):
                hints = records_format.hints if records_format else {'encoding': 'UTF8'}
                encoding = hints['encoding']
                raise TypeError("Please try again using a different encoding--"
                                f"{encoding} did not work")
            elif records_format is None:
                raise TypeError("Please try again and specify a particular records format--"
                                "assuming delimited with UTF-8 encoding did not work")
            else:
                raise

    def move_to_records_directory(self,
                                  records_directory: RecordsDirectory,
                                  records_format: BaseRecordsFormat,
                                  processing_instructions: ProcessingInstructions) -> MoveResult:

        def filename_from_url(url: str) -> str:
            return url[url.rfind('/')+1:]

        if records_format != self.records_format:
            raise NotImplementedError(f"This directory can only accept {self.records_format}")
        url_details = records_directory.save_fileobjs(self.target_names_to_input_fileobjs,
                                                      records_schema=self.records_schema,
                                                      records_format=self.records_format)
        output_urls = {
            filename_from_url(url): url
            for url in url_details
        }
        return MoveResult(output_urls=output_urls, move_count=None)

    @contextmanager
    def to_dataframes_source(self,
                             processing_instructions: ProcessingInstructions) \
            -> Iterator['DataframesRecordsSource']:
        from .dataframes import DataframesRecordsSource  # noqa

        """Convert current source to a DataframeSource and present it in a context manager"""
        if not isinstance(self.records_format, DelimitedRecordsFormat):
            raise NotImplementedError("Teach me to convert from "
                                      f"{self.records_format.format_type} to dataframe")
        unhandled_hints = set(self.records_format.hints.keys())
        options = pandas_read_csv_options(self.records_format,
                                          unhandled_hints,
                                          processing_instructions)
        complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                    unhandled_hints,
                                    self.records_format.hints)
        logger.info(f"Loading CSV via Pandas with options: {options}")
        hints = self.records_format.hints
        fileobjs = list(self.target_names_to_input_fileobjs.values())
        single_fileobj: IO[bytes] = ConcatFiles(fileobjs)  # type: ignore
        target_fileobj: IO[Any] = single_fileobj
        text_fileobj = None
        if hints['compression'] is None:
            hint_encoding: str = hints['encoding']  # type: ignore
            python_encoding = python_encoding_from_hint[hint_encoding]
            text_fileobj = io.TextIOWrapper(single_fileobj, encoding=python_encoding)
            target_fileobj = text_fileobj
        try:
            num_fields = len(self.records_schema.fields)
            entries_per_chunk = 2000000
            if num_fields == 0:
                chunksize = entries_per_chunk
            else:
                chunksize = int(entries_per_chunk / num_fields)

            try:
                dfs = pd.read_csv(filepath_or_buffer=target_fileobj,
                                  iterator=True,
                                  chunksize=chunksize,
                                  **options)
            except pd.errors.EmptyDataError:
                dfs = [self.records_schema.to_empty_dataframe()]
            yield DataframesRecordsSource(dfs=dfs, records_schema=self.records_schema)
        finally:
            if text_fileobj is not None:
                text_fileobj.detach()

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.records_format})"



import os
import logging


def adjusted_log_level(default_log_level: int, app_name: str) -> int:
    log_level_str = os.environ.get(app_name.upper() + '_LOG_LEVEL', None)
    if log_level_str is None:
        return default_log_level
    return getattr(logging, log_level_str.upper())


def set_levels(app_name: str,
               default_app_log_level: int=logging.INFO,
               default_root_log_level: int=logging.INFO) -> None:
    """
    Set up standard job log levels.  Can be overridden by
    YOUR_JOB_NAME_LOG_LEVEL env variable set to INFO/WARNING/etc

    :param app_name: module name of application; job logger is mounted here
    :param default_app_log_level: level of logging from application (default INFO)
    :param default_root_log_level: level of logging from root (default INFO)
    """

    # Log all messages to stderr
    root_logger = logging.getLogger()
    root_logger.setLevel(adjusted_log_level(default_root_log_level, app_name))

    # Log app messages to job_log by default
    app_logger = logging.getLogger(app_name)
    app_logger.setLevel(adjusted_log_level(default_app_log_level, app_name))



import io
from typing import IO, List


class ConcatFiles(io.RawIOBase):
    _files: List[IO[bytes]]

    def __init__(self, files: List[IO[bytes]]) -> None:
        self._files = files

    def close(self) -> None:
        for f in self._files:
            f.close()
        self._files = []
        return super().close()

    def readable(self) -> bool:
        return True

    def readall(self) -> bytes:
        return b''.join(f.read(-1) for f in self._files)

    def read(self, size: int = -1) -> bytes:
        if size < 0:
            return self.readall()

        while len(self._files) > 0:
            chunk = self._files[0].read(size)
            # If we aren't getting any bytes from this stream, lets move on to the next stream
            if len(chunk) == 0:
                f = self._files.pop(0)
                f.close()
            else:
                return chunk

        return b''



from contextlib import contextmanager
import json
from typing import TypeVar, Iterator, IO, Any, Mapping, Optional, List, Union

V = TypeVar('V', bound='BaseDirectoryUrl')


# Adapted from
# https://github.com/python/cpython/blob/3.7/Lib/shutil.py
# but this one returns the size copied
def blcopyfileobj(fsrc: IO[bytes], fdst: IO[bytes], length=16*1024):
    """copy data from file-like object fsrc to file-like object fdst"""
    written = 0
    while 1:
        buf = fsrc.read(length)
        if not buf:
            break
        written = written + fdst.write(buf)
    return written


class BaseDirectoryUrl:
    scheme: str
    url: str

    def __init__(self, url: str, **kwargs) -> None:
        raise NotImplementedError()

    def directory_in_this_directory(self, directory_name: str) -> 'BaseDirectoryUrl':
        "Create a subdirectory within this file's current directory with the given name"
        raise NotImplementedError()

    def file_in_this_directory(self, filename: str) -> 'BaseFileUrl':
        "Return an entry in the directory."
        raise NotImplementedError()

    def files_in_directory(self) -> List['BaseFileUrl']:
        "Return entries in this directory."
        raise NotImplementedError()

    def purge_directory(self) -> None:
        "Delete all entries and subdirectories in the directory."
        raise NotImplementedError()

    def files_and_directories_in_directory(self) -> List[Union['BaseFileUrl', 'BaseDirectoryUrl']]:
        "Return all file and folder entires under the current location"
        raise NotImplementedError()

    def copy_to(self, other_loc: 'BaseDirectoryUrl') -> 'BaseDirectoryUrl':
        "Copy all entries to the specified directory and return it"
        for file_or_directory in self.files_and_directories_in_directory():
            if file_or_directory.is_directory():
                source_subdirectory: BaseDirectoryUrl = file_or_directory  # type: ignore
                other_subdirname = source_subdirectory.filename()
                other_subdirectory = other_loc.directory_in_this_directory(other_subdirname)
                source_subdirectory.copy_to(other_subdirectory)
            else:
                source_file: BaseFileUrl = file_or_directory  # type: ignore
                other_filename = source_file.filename()
                other_file = other_loc.file_in_this_directory(other_filename)
                source_file.copy_to(other_file)
        return other_loc

    def is_directory(self) -> bool:
        "Returns true"
        raise NotImplementedError()

    def containing_directory(self: V) -> V:
        "Parent directory of this directory"
        raise NotImplementedError()

    @contextmanager
    def temporary_directory(self: V) -> Iterator[V]:
        "Yields a temporary DirectoryUrl in current location"
        raise NotImplementedError()

    def filename(self) -> str:
        "Filename, stripped of any directory information"
        raise NotImplementedError()

    def files_matching_prefix(self, prefix: str) -> List['BaseFileUrl']:
        "entries in the directory that match the prefix in its name."
        return [f for f in self.files_in_directory() if f.filename().startswith(prefix)]

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.url})"


T = TypeVar('T', bound='BaseFileUrl')


class BaseFileUrl:
    scheme: str
    url: str

    def __init__(self, url: str, **kwargs) -> None:
        raise NotImplementedError()

    def directory_in_this_directory(self, directory_name: str) -> BaseDirectoryUrl:
        "Create a subdirectory within this file's current directory with the given name"
        raise NotImplementedError()

    def string_contents(self, encoding: str='utf-8') -> str:
        "Return the contents of this file as a string."
        with self.open() as f:
            return f.read().decode(encoding)

    def json_contents(self) -> Optional[Mapping[Any, Any]]:
        file_contents = self.string_contents()
        if len(file_contents) > 0:
            data = json.loads(file_contents)
        else:
            data = None
        return data

    def file_in_this_directory(self, filename: str) -> 'BaseFileUrl':
        "Find a given file within this directory."
        raise NotImplementedError()

    def upload_fileobj(self, fileobj: IO[bytes], mode: str = 'wb') -> int:
        "Put the contents of the fileobj (stream) into the current file and "
        "return the number of bytes written."
        with self.open(mode=mode) as local_file:
            return blcopyfileobj(fileobj, local_file)

    def download_fileobj(self, output_fileobj: IO[bytes]) -> None:
        with self.open() as f:
            return blcopyfileobj(f, output_fileobj)

    def store_string(self, contents: str) -> None:
        "Save the specified string to the file."
        with self.open(mode='wt') as f:
            f.write(contents)

    def rename_to(self, new: 'BaseFileUrl') -> 'BaseFileUrl':
        "Rename this file"
        raise NotImplementedError()

    def is_directory(self) -> bool:
        "Returns false"
        return False

    def concatenate_from(self, other_locs: List['BaseFileUrl']) -> Optional[int]:
        length = 0
        with self.open(mode='wb') as output_fileobj:
            for input_loc in other_locs:
                with input_loc.open(mode='rb') as input_fileobj:
                    length += blcopyfileobj(input_fileobj, output_fileobj)
        return length

    def copy_to(self, other_loc: 'BaseFileUrl') -> 'BaseFileUrl':
        "Copy to the specified file and return it"
        with self.open() as fileobj:
            other_loc.upload_fileobj(fileobj)
        return other_loc

    def containing_directory(self) -> BaseDirectoryUrl:
        "Returns the parent directory of this file."
        raise NotImplementedError()

    @contextmanager
    def temporary_directory(self) -> Iterator[BaseDirectoryUrl]:
        parent_dir_loc = self.containing_directory()
        with parent_dir_loc.temporary_directory() as dir:
            yield dir

    def filename(self) -> str:
        "Filename, stripped of any directory information"
        raise NotImplementedError()

    def open(self, mode: str = "rb") -> IO[Any]:
        "Open file with the given mode.  Raises FileNotFoundError if file does not "
        "exist in the directory."
        raise NotImplementedError()

    def wait_to_exist(self):
        "Returns true after the file exists--useful for eventually consistent stores (e.g., S3)"
        return True

    def exists(self) -> bool:
        try:
            with self.open():
                return True
        except FileNotFoundError:
            return False

    def delete(self) -> None:
        raise NotImplementedError(f"Please implement for {type(self).__name__}")

    def size(self) -> int:
        raise NotImplementedError(f"Please implement for {type(self).__name__}")

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.url})"



from pandas import Series, Index
from typing import Any, Type, TYPE_CHECKING
from .statistics import RecordsSchemaFieldStringStatistics
from ...processing_instructions import ProcessingInstructions
from .representation import RecordsSchemaFieldRepresentation
from .numpy import details_from_numpy_dtype
import numpy as np
if TYPE_CHECKING:
    from ..field import RecordsSchemaField  # noqa
    from ..schema import RecordsSchema  # noqa


def field_from_index(index: Index,
                     processing_instructions: ProcessingInstructions) -> 'RecordsSchemaField':

    from ..field import RecordsSchemaField  # noqa
    field_type, constraints = details_from_numpy_dtype(index.dtype, unique=True)
    representations = {
        'origin': RecordsSchemaFieldRepresentation.from_index(index)
    }

    return RecordsSchemaField(name=index.name,
                              field_type=field_type,
                              constraints=constraints,
                              statistics=None,  # call refine_from_dataframe() for stats
                              representations=representations)


def field_from_series(series: Series,
                      processing_instructions: ProcessingInstructions) -> 'RecordsSchemaField':
    from ..field import RecordsSchemaField  # noqa
    field_type, constraints = details_from_numpy_dtype(series.dtype, unique=False)
    representations = {
        'origin': RecordsSchemaFieldRepresentation.from_series(series)
    }

    return RecordsSchemaField(name=series.name,
                              field_type=field_type,
                              constraints=constraints,
                              statistics=None,  # call refine_from_dataframe() for stats
                              representations=representations)


def refine_field_from_series(field: 'RecordsSchemaField',
                             series: Series,
                             total_rows: int,
                             rows_sampled: int) -> None:
    from ..field import RecordsSchemaField  # noqa
    #
    # if the series is full of object types that aren't numpy
    # types that show up directly as `.dtype` already, we can find
    # that out.
    #
    unique_python_types = series.map(type).unique()
    if unique_python_types.size == 1:
        unique_python_type: Type[Any] = unique_python_types[0]
        field_type = field.python_type_to_field_type(unique_python_type)
        if field_type is not None:
            if RecordsSchemaField.is_more_specific_type(field_type, field.field_type):
                field.field_type = field_type

    if field.field_type == 'string':
        max_column_length = series.astype('str').map(len).max()
        if not np.isnan(max_column_length):
            statistics =\
                RecordsSchemaFieldStringStatistics(rows_sampled=rows_sampled,
                                                   total_rows=total_rows,
                                                   max_length_bytes=None,
                                                   max_length_chars=max_column_length)
            if field.statistics is None:
                field.statistics = statistics
            elif not isinstance(field.statistics, RecordsSchemaFieldStringStatistics):
                raise SyntaxError("Did not expect to see existing statistics "
                                  f"for string type: {field.statistics}")
            else:
                field.statistics.merge(statistics)



from ..driver import DBDriver
import logging
from ...records import RecordsSchema
from ...records.unload_plan import RecordsUnloadPlan
from ...records.load_plan import RecordsLoadPlan
from ...records.records_format import BaseRecordsFormat, ParquetRecordsFormat
from ...records.types import RecordsFormatType
from ...records.records_directory import RecordsDirectory
from ...utils.limits import INT64_MAX, INT64_MIN, FLOAT64_SIGNIFICAND_BITS, num_digits
import re
from typing import Union, Optional, List, Tuple, IO
from ...url.resolver import UrlResolver
import sqlalchemy
from .loader import BigQueryLoader

logger = logging.getLogger(__name__)


class BigQueryDBDriver(DBDriver):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine],
                 url_resolver: UrlResolver,
                 **kwargs) -> None:
        super().__init__(db)
        self._bigquery_loader = BigQueryLoader(db=self.db, url_resolver=url_resolver)

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> int:
        """Loads the data from the RecordsDirectory instance named
        'directory'.  Guarantees a manifest file named 'manifest' is
        written to the target directory pointing to the target
        records.

        Returns number of rows loaded (if database provides that
        info).

        """
        return self._bigquery_loader.load(schema=schema,
                                          table=table,
                                          load_plan=load_plan,
                                          directory=directory)

    def load_from_fileobj(self, schema: str, table: str,
                          load_plan: RecordsLoadPlan, fileobj: IO[bytes]) -> Optional[int]:
        return self._bigquery_loader.load_from_fileobj(schema=schema,
                                                       table=table,
                                                       load_plan=load_plan,
                                                       fileobj=fileobj)

    def can_load_from_fileobjs(self):
        return True

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        return self._bigquery_loader.can_load_this_format(source_records_format)

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return self._bigquery_loader.known_supported_records_formats_for_load()

    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> int:
        """Writes table specified to the RecordsDirectory instance named 'directory'
        per the UnloadPlan named 'unload_plan'.  Guarantees a manifest
        file named 'manifest' is written to the target directory pointing
        to the target records.  After ensuring other records-related
        metadata is provided, you can call directory.finalize_manifest()
        to move this to the final location of _manifest, signaling to
        readers that the Records directory is finalized.

        Returns number of rows unloaded.
        """
        raise NotImplementedError(f"unload not implemented for this database type")

    def can_unload_this_format(self, target_records_format: BaseRecordsFormat) -> bool:
        return False

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        return []

    def best_records_format_variant(self, records_format_type: RecordsFormatType) ->\
            Optional[str]:
        if records_format_type == 'delimited':
            return 'bigquery'
        else:
            return None

    def type_for_date_plus_time(self, has_tz: bool=False) -> sqlalchemy.sql.sqltypes.DateTime:
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types
        if has_tz:
            # pybigquery doesn't set the timezone flag :(
            return sqlalchemy.sql.sqltypes.TIMESTAMP()
        else:
            return sqlalchemy.sql.sqltypes.DATETIME()

    def make_column_name_valid(self, colname: str) -> str:
        # https://cloud.google.com/bigquery/docs/schemas#column_names

        # replace anything which isn't a letter or number with '_',
        # which is the only valid symbol
        return re.sub(r'[^A-Za-z0-9]', '_', colname)

    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        # Only one int type for BigQuery, represented as sqlalchemy.types.Integer...
        return (INT64_MIN, INT64_MAX)

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        # Only one float type for BigQuery, represented as sqlalchemy.types.Float...
        return (64, FLOAT64_SIGNIFICAND_BITS)

    def fixed_point_constraints(self,
                                type_: sqlalchemy.types.Numeric) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.sql.sqltypes.DECIMAL):
            return (38, 9)
        else:
            logger.warning(f"Don't know how to handle unexpected BigQuery type {type(type_)}")
            return None

    def type_for_fixed_point(self,
                             precision: int,
                             scale: int) -> sqlalchemy.sql.sqltypes.Numeric:
        # BigQuery NUMERIC() type takes no arguments and supports 38
        # digits of precision, of which 9 digits are scale.
        if precision > 38 or scale > 9:
            logger.warning('Using BigQuery FLOAT64 type to represent '
                           f'NUMERIC({precision},{scale}))')
            return sqlalchemy.types.Float()
        return sqlalchemy.sql.sqltypes.Numeric()

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit the given min and max integer values"""

        if min_value is not None and max_value is not None:
            if min_value >= INT64_MIN and max_value <= INT64_MAX:
                # BigQuery has only 64-bit integers
                return sqlalchemy.sql.sqltypes.Integer()
            else:
                num_digits_min = num_digits(min_value)
                num_digits_max = num_digits(max_value)
                digit_count = max(num_digits_min, num_digits_max)
                return self.type_for_fixed_point(precision=digit_count,
                                                 scale=0)
        return super().type_for_integer(min_value, max_value)

    def tweak_records_schema_for_load(self,
                                      records_schema: RecordsSchema,
                                      records_format: BaseRecordsFormat) -> RecordsSchema:
        if isinstance(records_format, ParquetRecordsFormat):
            # According to Google, "DATETIME is not supported for
            # uploading from Parquet" -
            # https://github.com/googleapis/google-cloud-python/issues/9996#issuecomment-572273407
            #
            # So we need to make sure we don't create any DATETIME
            # columns if we're loading from a Parquet file.
            return records_schema.convert_datetimes_to_datetimetz()
        else:
            return records_schema



from records_mover import job_context  # noqa
from .types import JsonSchema  # noqa
from .version import __version__  # noqa
from .job_context import pull_job_context, create_job_context, run_as_job  # noqa



import functools
from typing import Callable, TypeVar


A = TypeVar('A')
B = TypeVar('B')


# https://stackoverflow.com/questions/3012421/python-memoising-deferred-lookup-property-decorator
def lazyprop(fn: Callable[[A], B]) -> B:
    attr_name = '_lazy_' + fn.__name__

    # https://github.com/python/mypy/issues/1362
    @property  # type: ignore
    @functools.wraps(fn)
    def _lazyprop(self: A) -> B:
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazyprop  # type: ignore



from typing import Dict, Optional, Union, List, Mapping, TYPE_CHECKING

"""RecordsHints are described as part of the overall `records format
documentation
<https://github.com/bluelabsio/knowledge/blob/master/Engineering/Architecture/JobDataExchange/output-design.md#hints>`_
Their goal in life is to both describe in detail how to write a
records file (especially delimited files, which are particularly
thorny in practice).

See the :py:meth:`records_mover.records.records_format.RecordsFormat`
for the other details that are typically provided along with records
hints.
"""
RecordsHints = Dict[str, Optional[Union[bool, str]]]

if TYPE_CHECKING:
    from ..db import DBDriver  # noqa
    # https://docs.aws.amazon.com/redshift/latest/dg/r_CREATE_EXTERNAL_TABLE.html
    from mypy_extensions import TypedDict

    class RecordsManifestEntryMetadata(TypedDict):
        content_length: int

    class BaseRecordsManifestEntry(TypedDict):
        url: str
        mandatory: bool

    class LegacyRecordsManifestEntry(BaseRecordsManifestEntry, total=False):
        meta: RecordsManifestEntryMetadata

    class RecordsManifestEntryWithLength(BaseRecordsManifestEntry):
        meta: RecordsManifestEntryMetadata

    class LegacyRecordsManifest(TypedDict):
        entries: List[LegacyRecordsManifestEntry]

    class RecordsManifestWithLength(TypedDict):
        entries: List[RecordsManifestEntryWithLength]

    class UrlDetailsEntry(TypedDict):
        content_length: int

    Url = str
    UrlDetails = Dict[Url, UrlDetailsEntry]

    from typing_extensions import Literal  # noqa

    from mypy_extensions import TypedDict
    BootstrappingRecordsHints = TypedDict('BootstrappingRecordsHints',
                                          {
                                              'quoting': Optional[str],
                                              'header-row': bool,
                                              'field-delimiter': str,
                                              'encoding': str,
                                              'escape': Optional[str],
                                              'compression': Optional[str],
                                          },
                                          total=False)

    RecordsFormatType = Literal['delimited', 'parquet']
else:
    RecordsManifestEntryMetadata = Mapping[str, int]
    LegacyRecordsManifestEntry = Mapping[str, Union[str, bool, int, RecordsManifestEntryMetadata]]
    RecordsManifestEntryWithLength =\
        Mapping[str, Union[str, bool, int, RecordsManifestEntryMetadata]]
    LegacyRecordsManifest = Mapping[str, List[LegacyRecordsManifestEntry]]
    RecordsManifestWithLength = Mapping[str, List[RecordsManifestEntryWithLength]]
    UrlDetailsEntry = Dict[str, int]
    Url = str
    UrlDetails = Dict[Url, UrlDetailsEntry]

    BootstrappingRecordsHints = RecordsHints

    RecordsFormatType = str



import sys
import subprocess
from ..base_job_context import BaseJobContext, JsonSchema
from .job_config_schema_as_args_parser import JobConfigSchemaAsArgsParser
from typing import Optional, Sequence
from ..creds.base_creds import BaseCreds
from ..creds.creds_via_lastpass import CredsViaLastPass


class CLIJobContext(BaseJobContext):
    __creds_via_lastpass_value: Optional[CredsViaLastPass]

    def __init__(self,
                 name: str,
                 default_db_creds_name: Optional[str],
                 default_aws_creds_name: Optional[str],
                 config_json_schema: Optional[JsonSchema],
                 scratch_s3_url=None,
                 args: Sequence[str]=sys.argv[1:]) -> None:
        super().__init__(name=name,
                         default_db_creds_name=default_db_creds_name,
                         default_aws_creds_name=default_aws_creds_name,
                         config_json_schema=config_json_schema,
                         scratch_s3_url=scratch_s3_url)
        if self._config_json_schema is not None:
            self.request_config =\
                JobConfigSchemaAsArgsParser.from_description(self._config_json_schema,
                                                             name).parse(args)
        else:
            self.request_config = {}
        self.__creds_via_lastpass_value = None

    @property
    def creds(self) -> BaseCreds:
        return self.creds_via_lastpass

    @property
    def creds_via_lastpass(self) -> CredsViaLastPass:
        if self.__creds_via_lastpass_value is None:
            self.__creds_via_lastpass_value = CredsViaLastPass()
        return self.__creds_via_lastpass_value

    @property
    def _scratch_s3_url(self) -> Optional[str]:
        if super()._scratch_s3_url is None:
            # check_output() is typed as returning Any in typeshed
            # - depends on encoding argument...
            try:
                self._scratch_s3_url_value: Optional[str] =\
                    subprocess.check_output("scratch-s3-url").decode('ASCII').rstrip()
            except FileNotFoundError:
                self._scratch_s3_url_value = None

        return self._scratch_s3_url_value



from .constraints import RecordsSchemaFieldConstraints
import sqlalchemy
from typing import Optional, Tuple, cast, TYPE_CHECKING
if TYPE_CHECKING:
    from .constraints import FieldStringConstraintsDict  # noqa
    from records_mover.db import DBDriver  # noqa


class RecordsSchemaFieldStringConstraints(RecordsSchemaFieldConstraints):
    def __init__(self,
                 required: bool,
                 unique: Optional[bool],
                 max_length_bytes: Optional[int],
                 max_length_chars: Optional[int]):
        super().__init__(required=required, unique=unique)
        self.max_length_bytes = max_length_bytes
        self.max_length_chars = max_length_chars

    @staticmethod
    def string_type_constraints(type_: sqlalchemy.types.String,
                                driver: 'DBDriver') ->\
            Tuple[Optional[int], Optional[int]]:
        max_length_bytes = None
        max_length_chars = None
        if driver.varchar_length_is_in_chars():
            max_length_chars = type_.length
        else:
            max_length_bytes = type_.length
        return (max_length_bytes, max_length_chars)

    @staticmethod
    def from_sqlalchemy_type(required: bool,
                             unique: Optional[bool],
                             type_: sqlalchemy.types.TypeEngine,
                             driver: 'DBDriver') -> 'RecordsSchemaFieldStringConstraints':
        if not isinstance(type_, sqlalchemy.types.String):
            raise TypeError(f"Unexpected column type: {type_}")
        max_length_bytes, max_length_chars =\
            RecordsSchemaFieldStringConstraints.string_type_constraints(type_,
                                                                        driver)
        return RecordsSchemaFieldStringConstraints(required=required,
                                                   unique=unique,
                                                   max_length_bytes=max_length_bytes,
                                                   max_length_chars=max_length_chars)

    def to_data(self) -> 'FieldStringConstraintsDict':
        raw_out = super().to_data()
        out = cast('FieldStringConstraintsDict', raw_out)
        if self.max_length_bytes is not None:
            out['max_length_bytes'] = self.max_length_bytes
        if self.max_length_chars is not None:
            out['max_length_chars'] = self.max_length_chars
        return out



from abc import ABCMeta, abstractmethod
from ..records_directory import RecordsDirectory
from ..results import MoveResult
from ..records_format import BaseRecordsFormat
from ..processing_instructions import ProcessingInstructions
from contextlib import contextmanager
from typing import Iterator, List
import itertools
from typing import TYPE_CHECKING, Optional
import logging
if TYPE_CHECKING:
    # http://mypy.readthedocs.io/en/latest/common_issues.html#import-cycles
    from .fileobjs import FileobjsSource  # noqa
    from .dataframes import DataframesRecordsSource  # noqa
    from ..targets import base as targets_base  # noqa

logger = logging.getLogger(__name__)


class RecordsSource(metaclass=ABCMeta):
    """These are possible abstract base classes you can implement --
    implement all which can be done efficiently for your
    RecordsSource.  See move() in mover.py to see how these are used
    or to debug a particular records move scenario).  None are
    required, but a more fully featured implementation is more likely
    to work with the current move() algorithms and allow moving
    records to a wider range of RecordsTarget classes."""

    def __str__(self) -> str:
        return type(self).__name__

    def validate(self) -> None:
        """Raise exception if this source as configured is not valid in some
        way (e.g., table doesn't exist, directory contains no files, etc)"""
        pass


class NegotiatesRecordsFormat(RecordsSource, metaclass=ABCMeta):
    def compatible_format(self,
                          records_target: 'targets_base.NegotiatesRecordsFormat')\
            -> Optional[BaseRecordsFormat]:
        source_formats = self.known_supported_records_formats()
        target_formats = records_target.known_supported_records_formats()
        compatible_format = None
        ranked_candidates = list(itertools.zip_longest(source_formats, target_formats))
        # Look at the candidates in order of their appearance, so if
        # there's a suboptimal records format, those can be put at the
        # end of the lists returned by
        # known_supported_records_formats() and the more preferred
        # ones can be put at the beginning.
        for source_candidate, target_candidate in ranked_candidates:
            if source_candidate is not None:
                if records_target.can_move_from_this_format(source_candidate):
                    compatible_format = source_candidate
                    break
            if target_candidate is not None:
                if self.can_move_to_this_format(target_candidate):
                    compatible_format = target_candidate
                    break
        if compatible_format is None:
            logger.warning(f"Mover: {self} is known to handle {source_formats} but "
                           f"is not able to directly export to {records_target}, "
                           f"which is known to handle {target_formats}")
        return compatible_format

    def has_compatible_format(self,
                              records_target: 'targets_base.NegotiatesRecordsFormat')\
            -> bool:
        return self.compatible_format(records_target) is not None

    @abstractmethod
    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        pass

    @abstractmethod
    def can_move_to_this_format(self,
                                target_records_format: BaseRecordsFormat) -> bool:
        """Return true if writing the specified format satisfies our format
        needs"""
        pass


class SupportsRecordsDirectory(RecordsSource, metaclass=ABCMeta):
    records_format: BaseRecordsFormat

    @abstractmethod
    def records_directory(self) -> RecordsDirectory:
        """Represent the current source as a RecordsDirectory object.
        Note that this is not a context manager, so don't implement
        this method by creating things which need cleanup (e.g.,
        creating a temporary directory).  Please see
        to_fileobjs_source() for that situation."""
        pass


class SupportsMoveToRecordsDirectory(NegotiatesRecordsFormat, metaclass=ABCMeta):
    @abstractmethod
    def move_to_records_directory(self,
                                  records_directory: RecordsDirectory,
                                  records_format: BaseRecordsFormat,
                                  processing_instructions: ProcessingInstructions) -> MoveResult:
        """Given an empty RecordsDirectory object, a RecordsFormat object
        representing the goal representation, and
        ProcessingInstructions, move the records into the target
        directory while converting their format."""
        pass


class SupportsToFileobjsSource(RecordsSource, metaclass=ABCMeta):
    @abstractmethod
    @contextmanager
    def to_fileobjs_source(self,
                           processing_instructions: ProcessingInstructions,
                           records_format_if_possible: Optional[BaseRecordsFormat]=None)\
            -> Iterator['FileobjsSource']:
        """Convert current source to a FileObjsSource and present it in a context manager.
        If there's no native records format, prefer 'records_format_if_possible' if provided
        and the source is capable of exporting to that format."""
        pass


class SupportsToDataframesSource(RecordsSource, metaclass=ABCMeta):
    @abstractmethod
    @contextmanager
    def to_dataframes_source(self,
                             processing_instructions: ProcessingInstructions) \
            -> Iterator['DataframesRecordsSource']:
        """Convert current source to a DataframeSource and present it in a context manager"""
        pass



from abc import abstractproperty, ABCMeta
import tempfile
import os
import jsonschema
from .database import db_engine, db_facts_from_env
from sqlalchemy.engine import Engine
import boto3
from .records.records import Records
from .db import db_driver, DBDriver
from .url.base import BaseFileUrl, BaseDirectoryUrl
import logging
import sqlalchemy
from typing import Dict, Any, Union, Optional
from .types import JsonSchema
from .creds.base_creds import BaseCreds
from .db.connect import engine_from_db_facts
from .url.resolver import UrlResolver
from db_facts.db_facts_types import DBFacts

RequestConfig = Dict[str, Any]


class BaseJobContext(metaclass=ABCMeta):
    __request_config: RequestConfig
    _scratch_s3_url_value: Optional[str]
    temp_dir: str

    def __init__(self,
                 name: str,
                 default_db_creds_name: Optional[str],
                 default_aws_creds_name: Optional[str],
                 config_json_schema: Optional[JsonSchema],
                 scratch_s3_url: Optional[str] = None) -> None:
        self.logger = logging.getLogger(name + '.__job__')
        self._default_db_creds_name = default_db_creds_name
        self._default_aws_creds_name = default_aws_creds_name
        # we must instantiate the temp dir alone and save it to self, otherwise
        # it will be garbage collected and the dir will be deleted
        self.temp_dir_obj = tempfile.TemporaryDirectory(prefix='records_mover_base_job_context')
        self.temp_dir = self.temp_dir_obj.name + '/'
        self._scratch_s3_url_value = scratch_s3_url
        self._config_json_schema = config_json_schema
        self.url_resolver = UrlResolver(boto3_session=self._boto3_session())

    def cleanup(self) -> None:
        self.temp_dir_obj.cleanup()

    @abstractproperty
    def creds(self) -> BaseCreds:
        raise NotImplementedError

    def get_default_db_engine(self) -> Engine:
        if self._default_db_creds_name is None:
            return db_engine(self)
        else:
            return self.get_db_engine(self._default_db_creds_name)

    def get_default_db_facts(self) -> DBFacts:
        if self._default_db_creds_name is None:
            return db_facts_from_env()
        else:
            return self.creds.db_facts(self._default_db_creds_name)

    def get_db_engine(self, db_creds_name: str, creds_provider: Optional[BaseCreds]=None) -> Engine:
        if creds_provider is None:
            creds_provider = self.creds
        db_facts = creds_provider.db_facts(db_creds_name)
        return engine_from_db_facts(db_facts)

    def db_driver(self, db: Union[sqlalchemy.engine.Engine,
                                  sqlalchemy.engine.Connection]) -> DBDriver:
        s3_temp_base_loc = None
        if self._scratch_s3_url is not None:
            s3_temp_base_loc = self.directory_url(self._scratch_s3_url)
        return db_driver(db=db,
                         s3_temp_base_loc=s3_temp_base_loc,
                         url_resolver=self.url_resolver)

    def file_url(self, url: str) -> BaseFileUrl:
        return self.url_resolver.file_url(url)

    def directory_url(self, url: str) -> BaseDirectoryUrl:
        return self.url_resolver.directory_url(url)

    def _boto3_session(self) -> boto3.session.Session:
        if self._default_aws_creds_name is None:
            return boto3.session.Session()
        else:
            return self.creds.boto3_session(self._default_aws_creds_name)

    @property
    def records(self) -> Records:
        return Records(db_driver=self.db_driver,
                       url_resolver=self.url_resolver,
                       logger=self.logger)

    def _validate_config(self,
                         config_json_schema: JsonSchema,
                         request_config:
                         RequestConfig) -> None:
        if config_json_schema is not None:
            jsonschema.validate(request_config, config_json_schema)

    @property
    def _scratch_s3_url(self) -> Optional[str]:
        if self._scratch_s3_url_value is None:
            if "SCRATCH_S3_URL" in os.environ:
                self._scratch_s3_url_value = os.environ["SCRATCH_S3_URL"]
        if self._scratch_s3_url_value is not None:
            if not self._scratch_s3_url_value.endswith('/'):
                raise ValueError("Please provide a directory name - "
                                 f"URL should end with '/': {self._scratch_s3_url_value}")
        return self._scratch_s3_url_value

    @property
    def request_config(self) -> RequestConfig:
        return self.__request_config

    @request_config.setter
    def request_config(self, value: RequestConfig) -> None:
        if self._config_json_schema is not None:
            self._validate_config(self._config_json_schema, value)
        self.__request_config = value



class LoadUnloadError(Exception):
    pass


class NoTemporaryBucketConfiguration(LoadUnloadError):
    pass


class LoadError(LoadUnloadError):
    pass


class UnloadError(LoadUnloadError):
    pass


class S3ExportNotSupported(UnloadError):
    pass


class DatabaseDoesNotSupportS3Export(S3ExportNotSupported):
    pass


class CredsDoNotSupportS3Export(S3ExportNotSupported):
    pass


class CredsDoNotSupportS3Import(LoadError):
    pass



from contextlib import contextmanager
import os
import tempfile
from typing import Iterator


@contextmanager
def set_temp_dir(new_temp_dir: str) -> Iterator[None]:
    old_env = dict(os.environ)
    old_tempdir = tempfile.tempdir
    try:
        # https://docs.python.org/3/library/tempfile.html#tempfile.gettempdir
        env = os.environ.copy()
        env['TMPDIR'] = new_temp_dir
        env['TEMP'] = new_temp_dir
        env['TMP'] = new_temp_dir
        os.environ.update(env)

        tempfile.tempdir = new_temp_dir
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_env)
        tempfile.tempdir = old_tempdir



#!/usr/bin/env python

import os
from setuptools import setup, find_packages
from setuputils import TestCoverageRatchetCommand, MypyCoverageRatchetCommand, VerifyVersionCommand

gsheet_dependencies = [
    'google',
    'google_auth_httplib2',
    'google-api-python-client>=1.5.0,<1.6.0',
    'oauth2client>=2.0.2,<2.1.0',
    'PyOpenSSL'
]

__version__: str
# Read in and set version variable without the overhead/requirements
# of the rest of the package.
#
# https://milkr.io/kfei/5-common-patterns-to-version-your-Python-package/5
dir_path = os.path.dirname(os.path.realpath(__file__))
exec(open(os.path.join(dir_path, 'records_mover', 'version.py')).read())

setup(name='records-mover',
      version=__version__, # read right above  # noqa
      description=('Library and CLI to move relational data from one place to another - '
                   'DBs/CSV/gsheets/dataframes/...'),
      author='Vince Broz',
      author_email='vince.broz@bluelabs.com',
      packages=find_packages(),
      package_data={
          'records_mover': ['py.typed']
      },
      install_requires=[
          'boto>=2,<3', 'boto3',
          'jsonschema', 'timeout_decorator', 'awscli',
          'PyYAML', 'psycopg2',
          # sqlalchemy-vertica-python 0.5.5 introduced
          # https://github.com/bluelabsio/sqlalchemy-vertica-python/pull/7
          # which fixed a bug pulling schema information from Vertica
          'sqlalchemy-vertica-python>=0.5.5,<0.6',
          # sqlalchemy-redshift 0.7.7 introduced support for Parquet
          # in UNLOAD
          'sqlalchemy-redshift>=0.7.7',
          'pybigquery',
          'sqlalchemy',
          # Not sure how/if interface will change in db-facts, so
          # let's be conservative about what we're specifying for now.
          'db-facts>=2.15.3,<3',
          'odictliteral',
          # we rely on exception types from smart_open,
          # which seem to change in feature releases
          # without a major version bump
          'smart_open>=1.8.4,<1.9.0',
          'chardet',
          's3-concat>=0.1.7,<0.2'
          'tenacity>=6<7'
      ],
      extras_require={
          'gsheets': gsheet_dependencies,
          'movercli': gsheet_dependencies + ['typing_inspect', 'docstring_parser',
                                             'pandas<2',
                                             'pyarrow'],
      },
      entry_points={
          'console_scripts': 'mvrec = records_mover.records.cli:main',
      },
      cmdclass={
          'coverage_ratchet': TestCoverageRatchetCommand,
          'mypy_ratchet': MypyCoverageRatchetCommand,
          'verify': VerifyVersionCommand,
      })



from .base import SupportsRecordsDirectory
from records_mover.db.quoting import quote_schema_and_table
from ...db import DBDriver
from ...url.resolver import UrlResolver
from ...url import BaseDirectoryUrl
from sqlalchemy.engine import Engine, Connection
from ..records_directory import RecordsDirectory
from ..records_format import ParquetRecordsFormat
from sqlalchemy.schema import CreateTable, MetaData, Table
from ..existing_table_handling import ExistingTableHandling
from typing import Optional, Callable, Union
import logging
import sqlalchemy


logger = logging.getLogger(__name__)


class SpectrumRecordsTarget(SupportsRecordsDirectory):
    def __init__(self,
                 schema_name: str,
                 table_name: str,
                 db_engine: Engine,
                 db_driver: Callable[[Union[Engine, Connection]], DBDriver],
                 url_resolver: UrlResolver,
                 spectrum_base_url: Optional[str],
                 spectrum_rdir_url: Optional[str],
                 existing_table_handling: ExistingTableHandling=
                 ExistingTableHandling.TRUNCATE_AND_OVERWRITE) -> None:
        self.db = db_engine
        self.driver = db_driver(db_engine)
        self.schema_name = schema_name
        self.table_name = table_name
        self.url_resolver = url_resolver
        self.existing_table_handling = existing_table_handling
        self.output_loc = self.calculate_output_url(spectrum_base_url=spectrum_base_url,
                                                    spectrum_rdir_url=spectrum_rdir_url)

        if not self.output_loc.scheme == 's3':
            raise ValueError("Records URL for Redshift Spectrum must be in S3")
        self.records_format = ParquetRecordsFormat()

    def calculate_output_url(self,
                             spectrum_base_url: Optional[str],
                             spectrum_rdir_url: Optional[str]) -> BaseDirectoryUrl:
        if spectrum_rdir_url is not None:
            return self.url_resolver.directory_url(spectrum_rdir_url)
        elif spectrum_base_url is not None:
            if not spectrum_base_url.endswith('/'):
                raise ValueError("Please provide a directory name - "
                                 f"URL should end with '/': {spectrum_base_url}")
            return\
                self.url_resolver.directory_url(spectrum_base_url).\
                directory_in_this_directory('data').\
                directory_in_this_directory(self.schema_name).\
                directory_in_this_directory(self.table_name)
        else:
            raise TypeError("Please specify either spectrum_base_url or spectrum_s3_rdir_url")

    def prep_bucket(self) -> None:
        # These are modes we may want to try to make work in the
        # future, but Redshift doesn't support them directly today so
        # we'd need to do some lower level things (e.g., truncating
        # the actual Parquet files)
        if self.existing_table_handling in [
                ExistingTableHandling.TRUNCATE_AND_OVERWRITE,
                ExistingTableHandling.DROP_AND_RECREATE,
                ExistingTableHandling.DELETE_AND_OVERWRITE
        ]:
            if self.existing_table_handling == ExistingTableHandling.DELETE_AND_OVERWRITE:
                logger.warning('Redshift Spectrum does not support transactional delete.')
            if self.existing_table_handling == ExistingTableHandling.DROP_AND_RECREATE:
                schema_and_table: str = quote_schema_and_table(self.db,
                                                               self.schema_name,
                                                               self.table_name)
                logger.info(f"Dropping external table {schema_and_table}...")
                with self.db.connect() as cursor:
                    # See below note about fix from Spectrify
                    cursor.execution_options(isolation_level='AUTOCOMMIT')
                    cursor.execute(f"DROP TABLE IF EXISTS {schema_and_table}")

            logger.info(f"Deleting files in {self.output_loc}...")
            self.output_loc.purge_directory()
        elif self.existing_table_handling == ExistingTableHandling.APPEND:
            raise NotImplementedError('APPEND mode not yet supported')
        else:
            raise NotImplementedError(f'Teach me how to handle {self.existing_table_handling}')

    def pre_load_hook(self) -> None:
        self.prep_bucket()

    def records_directory(self) -> RecordsDirectory:
        return RecordsDirectory(self.output_loc)

    def post_load_hook(self, num_rows_loaded: Optional[int]) -> None:
        if self.existing_table_handling != ExistingTableHandling.DROP_AND_RECREATE:
            return

        directory = self.records_directory()
        records_schema = directory.load_schema_json_obj()
        records_manifest = directory.get_manifest()
        assert records_manifest is not None  # just written before this was called
        assert records_schema is not None  # just written before this was called

        meta = MetaData()
        columns = [f.to_sqlalchemy_column(self.driver) for f in records_schema.fields]
        for column in columns:
            if isinstance(column.type, sqlalchemy.sql.sqltypes.Numeric) and column.type.asdecimal:
                # https://app.asana.com/0/1128138765527694/1139365728890234
                raise NotImplementedError("Teach me how to write a NUMERIC to Redshift Spectrum "
                                          f"(column name: {column})")
            if isinstance(column.type, sqlalchemy.sql.sqltypes.DateTime) and column.type.timezone:
                # https://app.asana.com/0/53283930106309/1136051640283464
                raise NotImplementedError("Teach me how to write a datetimetz to Redshift Spectrum "
                                          f"({column})")

        table = Table(self.table_name, meta,
                      prefixes=["EXTERNAL"],
                      *columns,
                      schema=self.schema_name)
        schema_sql = str(CreateTable(table, bind=self.driver.db_engine))
        table_properties_clause = ''
        if num_rows_loaded is not None:
            table_properties_clause = f"\nTABLE PROPERTIES ('numRows'='{num_rows_loaded}')"
        else:
            # If the move from the source doesn't give us a row count
            # for use as statistics by the Redshift query optimizer,
            # we could potentially gather it manually:
            #
            # https://docs.aws.amazon.com/redshift/latest/dg/r_ALTER_TABLE_external-table.html
            #
            # https://app.asana.com/0/53283930106309/1140172717234001
            pass
        storage_clause = "STORED AS PARQUET\n"
        location_clause = f"LOCATION '{self.output_loc.url}_manifest'\n"
        schema_sql = schema_sql + storage_clause + location_clause + table_properties_clause

        logger.info(schema_sql)
        with self.driver.db_engine.connect() as cursor:
            # without autocommit, we get "CREATE EXTERNAL TABLE cannot
            # run inside a transaction block"
            #
            # Fix from Spectrify:
            #
            # https://github.com/hellonarrativ/spectrify/issues/12
            # https://github.com/hellonarrativ/spectrify/pull/13/files#diff-d604d159eb266aacefeebac327bacd26R57

            cursor.execution_options(isolation_level='AUTOCOMMIT')
            cursor.execute(schema_sql)



import sqlalchemy
from typing import Callable, Union, Optional
from sqlalchemy.sql import text


select_load_errors = """
        select d.query, substring(d.filename,14,20),
        d.line_number as line,
        substring(d.value,1,16) as value,
        substring(le.err_reason,1,100) as err_reason
        from stl_loaderror_detail d, stl_load_errors le
        where d.query = le.query
        and d.query = :last_copy_id;
        """


def schema_sql_from_admin_views(schema: str,
                                table: str,
                                db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                                text: Callable[[str], sqlalchemy.sql.expression.TextClause]=text)\
        -> Optional[str]:
    # The default behavior in current sqlalchemy driver seems to
    # map "DOUBLE PRECISION" to "DOUBLE_PRECISION", so lean back
    # on admin views for now:
    sql = text("""
SELECT ddl
FROM admin.v_generate_tbl_ddl
WHERE schemaname = :schema_name
AND tablename = :table_name
""")
    result = db.execute(sql, schema_name=schema, table_name=table).fetchall()
    if len(result) == 0:
        return None
    else:
        # First line is a commented-out drop table, don't need that.
        return "\n".join(map(lambda r: r['ddl'], result[1:]))



import sys
import logging


stdout_handler = logging.StreamHandler(stream=sys.stdout)
formatter = logging.Formatter('%(asctime)s - %(message)s', '%H:%M:%S')
stdout_handler.setFormatter(formatter)


def basic_config() -> None:
    """Set up standard job logging for CLI use.

    Sends any log statements to stdout, with a relatively short
    timestamp beforehand (helpful for timing long-running DS work).
    """
    root_logger = logging.getLogger()
    root_logger.addHandler(stdout_handler)



"""Create and run jobs to convert between different sources and targets"""
from records_mover.job_context import create_job_context
from ..processing_instructions import ProcessingInstructions
from ..results import MoveResult
from ...types import JobConfig
from .config import config_to_args


def run_records_mover_job(source_method_name: str,
                          target_method_name: str,
                          job_name: str,
                          config: JobConfig) -> MoveResult:
    with create_job_context(job_name) as job_context:
        try:
            source_method = getattr(job_context.records.sources, source_method_name)
            target_method = getattr(job_context.records.targets, target_method_name)
            job_context.logger.info('Starting...')
            logger = job_context.logger
            logger.debug("Initialized class!")
            records = job_context.records

            source_kwargs = config_to_args(config=config['source'],
                                           method=source_method,
                                           job_context=job_context)
            target_kwargs = config_to_args(config=config['target'],
                                           method=target_method,
                                           job_context=job_context)
            pi_config = {k: config[k] for k in config if k not in ['source', 'target', 'func']}
            pi_kwargs = config_to_args(pi_config,
                                       method=ProcessingInstructions,
                                       job_context=job_context)
            processing_instructions = ProcessingInstructions(**pi_kwargs)

            source = source_method(**source_kwargs)
            target = target_method(**target_kwargs)
            return records.move(source, target, processing_instructions)
        except Exception as e:
            job_context.logger.error(e)
            raise



from pandas import DataFrame
from .to_csv_options import pandas_to_csv_options  # noqa
from .read_csv_options import pandas_read_csv_options  # noqa


def _lowercase_column_names(df: DataFrame) -> DataFrame:
    df.columns = map(str.lower, df.columns)
    return df


def prep_df_for_loading(df: DataFrame) -> DataFrame:
    return _lowercase_column_names(df)



"""Defines exceptions used by records module."""


class RecordsException(Exception):
    """Base class for all records system exceptions."""

    pass


class RecordsFolderNonEmptyException(RecordsException):
    """Raised if trying to write records to a non-empty target directory."""

    pass
